import pandas as pd
import pyspark.sql.functions as F


def bootstrap(df, iterations, reporting_freq=10, strata=[],
              treatment='variant', control='control'):
    def munge(s):
        agg = s.groupby(["experiment_branch"] + strata).mean()
        agg = (
          pd.melt(agg.reset_index(), id_vars=["experiment_branch"] + strata)
          .pivot_table(
                    index=strata + ['variable'],
                    columns='experiment_branch',
                    values='value').reset_index()
          )
        agg['change'] = (agg[treatment] - agg[control]) / agg[control]
        return agg

    n = len(df)
    results = munge(df)
    results['iteration'] = -1  # true sample stats lie at -1
    for i in range(iterations):
        if i % 10 == 0:
            print(i, end=' ')
        s = df.sample(n, replace=True)
        sagg = munge(s)
        sagg['iteration'] = i
        results = pd.concat([sagg, results])
    return results


measures = {
  'google_sap': 500,
  'sap': 500,
  'google_organic': 500,
  'organic': 500,
  'clicks': 100,
  'impressions': 300,
  'tab_open_event_count': 500,
  'uri_count': 1000
}


def cap_daily_measures(df, measures):
    for m in measures:
        col = F.col(m)
        cap = measures[m]
        df = df.withColumn(m, F.when(col > cap, cap).otherwise(col))
    return df


def percent_capped(df, measures):
    return(df
           .agg(*[(F.avg(
                  F.when(F.col(m) == measures[m], 1)
                   .otherwise(0))).alias(m)
                  for m in measures])
           ).toPandas().T



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from plotly.offline import plot
import plotly.graph_objs as go
from fbprophet.plot import add_changepoints_to_plot

from dscontrib.jmccrosky.forecast.utils import calcMAPE


def evaluateModel(model, data, endDate=None, title=None):
    model.fit(data["training"])
    holdout_period = model.make_future_dataframe(
        periods=len(data["holdout"]), include_history=False
    )
    holdout_forecast = model.predict(holdout_period)
    if endDate is None:
        periods = len(data["holdout"])
    else:
        periods = (endDate - data["training"].ds.max()).days
    all_period = model.make_future_dataframe(periods=periods, include_history=True)
    all_forecast = model.predict(all_period)
    text = '{}Holdout MAPE: {:,.2f}%'.format(
        "{} - ".format(title) if title is not None else "",
        calcMAPE(data["holdout"].y, holdout_forecast.yhat)
    )
    plotHTML = plot({"data": [
        go.Scatter(x=data["all"]['ds'], y=data["all"]['y'], name='y'),
        go.Scatter(x=all_forecast['ds'], y=all_forecast['yhat'], name='yhat'),
        go.Scatter(
            x=all_forecast['ds'], y=all_forecast['yhat_upper'], fill='tonexty',
            mode='none', name='upper'
        ),
        go.Scatter(
            x=all_forecast['ds'], y=all_forecast['yhat_lower'], fill='tonexty',
            mode='none', name='lower'
        ),
    ], "layout": go.Layout(title=text)}, output_type='div')
    plotProphet = model.plot(all_forecast)
    add_changepoints_to_plot(plotProphet.gca(), model, all_forecast)
    plotComponents = model.plot_components(all_forecast)
    return {
        "plot": plotHTML,
        "prophetplot": plotProphet,
        "plotcomponents": plotComponents
    }



"""
Luigi!

To get all the dependencies you need to use this, install dscontrib like:
    pip install dscontrib[luigi]

You'll also need to set up application default credentials:
https://docs.telemetry.mozilla.org/cookbooks/bigquery.html#from-client-sdks

You may find yourself in the following scenario:
* You have a bunch of .sql files, which may or may not depend on each other,
  that define some queries. You're probably interested in materializing some,
  but not necessarily all, of those queries directly into .csv.gz files in your
  working directory.
* You'd like to update your local .csv files update when you change one of your .sql
  files.
* You don't want to run more queries than necessary.
* You don't want to have to copy-paste things and hold the state of your query
  DAG in your head.

This is for you! This helps you use Luigi to manage the state of your workflow.
It works great for either simple or complex cases.

We don't need to worry too much about how Luigi works in order to use this,
but it's good to know that Luigi has a concept of _tasks_,
which have dependencies, take inputs, and produce outputs.

This file defines some different kinds of tasks. You declare your specific
workflow by inheriting from them.

An example workflow looks like:

### workflow.py

import luigi
from dscontrib.tdsmith.luigi import *

class ExtractPingsTable(TableFromStalableSql):
    # Run a SQL job and save the results to a BQ table.
    # This doesn't download any rows directly.
    # It's great for subsetting down to pings you need
    # so that they're faster to query later.
    query_filename = "extract_pings.sql"
    destination_table = "moz-fx-data-bq-data-science.tdsmith.extracted_pings"

class SummarizePingsTable(TableFromStalableSql):
    # Same thing, except this job depends on the extract job.
    query_filename = "summarize_pings.sql"
    destination_table = "moz-fx-data-bq-data-science.tdsmith.summarized_pings"

    # This is how we express that this job depends on the last one:
    def requires(self):
        return ExtractPingsTable()

class PingSummaryCsv(ExtractedCsvFromStalableTable):
    # This class keeps a local CSV file up to date with a BigQuery table.
    source_table = SummarizePings.destination_table
    storage_url = "gs://bucket/filename.csv.gz"
    output_filename = "summary.csv.gz"

    def requires(self):
        return SummarizePingsTable()

# We can also define other queries that write directly to a .csv.
# This is useful for small result sets.
class EnrollmentCsv(CsvFromStalableSql):
    query_filename = "enrollment.sql"
    output_filename = "enrollment.csv.gz"

# Finally, we can define a "parent" task that makes sure everything gets run.
# It should depend on all of your CSVs.
class MyWorkflow(luigi.Task):
    def requires(self):
        return [
            PingSummaryCsv(),
            EnrollmentCsv(),
        ]
###

Now, you can run this with:
    PYTHONPATH=$PWD python -m luigi --module workflow MyWorkflow --local-scheduler

The first time, it will run all your queries.
The second time, it won't run any of them, until you change the .sql files!
"""


import datetime as dt
from functools import partial
import json
from pathlib import Path
import logging

import attr
from google.cloud import bigquery as bq
from google.cloud import storage as storage
from google.api_core import exceptions as bq_ex
import luigi
from pytz import utc


__all__ = [
    "CsvFromStalableSql",
    "TableFromStalableSql",
    "ExtractedCsvFromStalableTable",
]


PROJECT = "moz-fx-data-bq-data-science"
logger = logging.getLogger("luigi-interface")


@attr.s
class BigQueryTarget(luigi.Target):
    table: bq.table.TableReference = attr.ib()
    client: bq.Client = attr.ib(default=attr.Factory(partial(bq.Client, PROJECT)))

    _table_object: bq.table.Table = attr.ib(default=None)

    @property
    def table_object(self):
        if self._table_object:
            return self._table_object
        try:
            self._table_object = self.client.get_table(self.table)
        except bq_ex.NotFound:
            self._table_object = None
        return self._table_object

    def exists(self):
        return self.table_object is not None


class CsvFromStalableSql(luigi.Task):
    """Return a small result set from an anonymous query."""

    @property
    def query_filename(self):
        raise NotImplementedError

    @property
    def output_filename(self):
        raise NotImplementedError

    query_parameters = None

    __client = None

    @property
    def client(self):
        self.__client = self.__client or bq.Client(PROJECT)
        return self.__client

    def run(self):
        query = Path(self.query_filename).read_text()

        query_parameters = []
        if self.query_parameters:
            query_parameters = [
                bq.query._query_param_from_api_repr(i)
                for i in json.loads(Path(self.query_parameters).read_text())
            ]

        job = self.client.query(
            query,
            bq.job.QueryJobConfig(
                query_parameters=query_parameters,
                use_legacy_sql=False,
            ),
        )
        df = job.to_dataframe()
        df.to_csv(self.output_filename, index=False)

    def output(self):
        return luigi.LocalTarget(self.output_filename)

    def complete(self):
        for task in luigi.task.flatten(self.requires()):
            if not task.complete():
                return False

        output_path = Path(self.output().path)
        if not output_path.exists():
            return False
        output_mtime = output_path.stat().st_mtime
        input_mtime = Path(self.query_filename).stat().st_mtime
        return input_mtime < output_mtime


class TableFromStalableSql(luigi.Task):
    @property
    def query_filename(self):
        raise NotImplementedError

    @property
    def destination_table(self):
        raise NotImplementedError

    query_parameters = None

    __client = None

    @property
    def client(self):
        self.__client = self.__client or bq.Client(PROJECT)
        return self.__client

    def run(self):
        logger.info("Running query")
        query = Path(self.query_filename).read_text()

        query_parameters = []
        if self.query_parameters:
            query_parameters = [
                bq.query._query_param_from_api_repr(i)
                for i in json.loads(Path(self.query_parameters).read_text())
            ]

        job = self.client.query(
                query,
                bq.job.QueryJobConfig(
                    destination=self.destination_table,
                    query_parameters=query_parameters,
                    use_legacy_sql=False,
                    write_disposition=bq.job.WriteDisposition.WRITE_TRUNCATE,
                )
        )
        job.result(max_results=1)

    def output(self):
        ref = bq.table.TableReference.from_string(self.destination_table)
        return BigQueryTarget(ref, self.client)

    def complete(self):
        for task in luigi.task.flatten(self.requires()):
            if not task.complete():
                return False

        table = self.output().table_object
        if table is None:
            logger.info("Destination table does not exist")
            return False

        query_mtime = Path(self.query_filename).stat().st_mtime
        query_mtime_dt = dt.datetime.fromtimestamp(query_mtime, utc)
        logger.info(f"Local: {query_mtime_dt} Remote: {table.modified}")
        return query_mtime_dt < table.modified


class ExtractedCsvFromStalableTable(luigi.Task):
    @property
    def source_table(self):
        raise NotImplementedError

    @property
    def output_filename(self):
        raise NotImplementedError

    @property
    def storage_url(self):
        """gs://bucket/filename.csv.gz"""
        raise NotImplementedError

    __client = None

    @property
    def client(self):
        self.__client = self.__client or bq.Client(PROJECT)
        return self.__client

    def run(self):
        job = self.client.extract_table(
            self.source_table,
            self.storage_url,
            job_config=bq.job.ExtractJobConfig(
                compression=bq.job.Compression.GZIP,
                destination_format=bq.job.DestinationFormat.CSV,
            )
        )
        job.result()

        storage_client = storage.client.Client(self.client.project)

        with open(self.output_filename, "wb") as f:
            storage_client.download_blob_to_file(self.storage_url, f)

    def output(self):
        return luigi.LocalTarget(self.output_filename)

    def complete(self):
        for task in luigi.task.flatten(self.requires()):
            if not task.complete():
                return False

        output = Path(self.output_filename)
        if not output.exists():
            return False
        output_mtime = dt.datetime.fromtimestamp(output.stat().st_mtime, utc)

        source_table_obj = self.client.get_table(self.source_table)
        return source_table_obj.modified < output_mtime



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from fbprophet import Prophet
import pandas as pd
from datetime import date, timedelta
from dscontrib.jmccrosky.forecast.utils import s2d


# Get easter dates
def getEasters(year):
    a = year % 19
    b = year // 100
    c = year % 100
    d = (19 * a + b - b // 4 - ((b - (b + 8) // 25 + 1) // 3) + 15) % 30
    e = (32 + 2 * (b % 4) + 2 * (c // 4) - d - (c % 4)) % 7
    f = d + e - 7 * ((a + 11 * d + 22 * e) // 451) + 114
    month = f // 31
    day = f % 31 + 1
    easter = date(year, month, day)
    return [("easter{}".format(x), easter + timedelta(days=x)) for x in range(-3, 2)]


# Get holidays dataframe in prophet's format
def getHolidays(years):
    easters = pd.DataFrame({
        'ds': [e[1] for i in years for e in getEasters(i)],
        'holiday': [e[0] for i in years for e in getEasters(i)],
        'lower_window': 0,
        'upper_window': 0,
    })
    return easters


def setupModels(years):
    models = {}
    models["desktop_global"] = Prophet(
        yearly_seasonality=20,
        changepoint_range=0.7,
        seasonality_mode='multiplicative',
        changepoint_prior_scale=0.015,
        seasonality_prior_scale=0.25,
        holidays=getHolidays(years)
    )
    models["nondesktop_global"] = Prophet()
    models["fxa_global"] = Prophet(
        changepoint_range=0.8,
        changepoint_prior_scale=0.02,
    )
    models["desktop_tier1"] = Prophet()
    models["nondesktop_tier1"] = Prophet()
    models["fxa_tier1"] = Prophet(
        changepoint_range=0.8,
        changepoint_prior_scale=0.02,
    )
    models["Fennec Android"] = Prophet(
        changepoint_prior_scale=0.0005,
        seasonality_prior_scale=0.001,
        seasonality_mode='multiplicative'
    )
    models["Focus iOS"] = Prophet(changepoint_prior_scale=0.0005)
    models["Focus Android"] = Prophet(changepoint_prior_scale=0.005)
    models["Fennec iOS"] = Prophet(
        changepoint_prior_scale=0.005,
        seasonality_prior_scale=0.001,
        seasonality_mode='multiplicative'
    )
    models["Fenix"] = Prophet(changepoint_prior_scale=0.0005)
    models["Firefox Lite"] = Prophet(changepoint_prior_scale=0.0005)
    models["FirefoxForFireTV"] = Prophet(
        changepoint_prior_scale=0.0005,
        seasonality_prior_scale=0.005,
        seasonality_mode='multiplicative',
        yearly_seasonality=True
    )
    models["FirefoxConnect"] = Prophet(changepoint_prior_scale=0.0005)
    models["nondesktop_nofire_global"] = Prophet(
        yearly_seasonality=20,
        changepoint_range=0.75,
        seasonality_mode='multiplicative',
        changepoint_prior_scale=0.008,
        seasonality_prior_scale=0.20,
        holidays=getHolidays(years)
    )
    models["nondesktop_nofire_tier1"] = Prophet(
        yearly_seasonality=20,
        changepoint_range=0.75,
        seasonality_mode='multiplicative',
        changepoint_prior_scale=0.008,
        seasonality_prior_scale=0.20,
        holidays=getHolidays(years)
    )
    models["nondesktop_nofire_global_2020"] = Prophet(
        yearly_seasonality=20,
        changepoint_range=0.75,
        seasonality_mode='multiplicative',
        changepoint_prior_scale=0.008,
        seasonality_prior_scale=0.0002,
        holidays=getHolidays(years)
    )
    models["nondesktop_nofire_tier1_2020"] = Prophet(
        yearly_seasonality=20,
        changepoint_range=0.75,
        seasonality_mode='multiplicative',
        changepoint_prior_scale=0.008,
        seasonality_prior_scale=0.0002,
        holidays=getHolidays(years)
    )
    return models


def dataFilter(data, product):
    startDates = {
        "desktop_global": s2d('2016-04-08'),
        "fxa_global": s2d('2018-03-20'),
        "fxa_tier1": s2d('2018-03-20'),
        "Fennec Android": s2d('2017-03-04'),
        "Focus iOS": s2d('2017-12-06'),
        "Focus Android": s2d('2017-07-17'),
        "Fennec iOS": s2d('2017-03-03'),
        "Fenix": s2d('2019-07-03'),
        "Firefox Lite": s2d('2017-03-04'),
        "FirefoxForFireTV": s2d('2018-02-04'),
        "FirefoxConnect": s2d('2018-10-10'),
        "nondesktop_nofire_global": s2d('2017-01-30'),
        "nondesktop_nofire_tier1": s2d('2017-01-30'),
        "nondesktop_nofire_global_2020": s2d('2017-01-30'),
        "nondesktop_nofire_tier1_2020": s2d('2017-01-30'),
    }

    anomalyDates = {
        "desktop_global": [s2d('2019-05-16'), s2d('2019-06-07')],
        "Focus Android": [s2d('2018-09-01'), s2d('2019-03-01')],
        "Fennec iOS": [s2d('2017-11-08'), s2d('2017-12-31')],
        "nondesktop_nofire_global_2020": [s2d('2017-11-10'), s2d('2018-03-11')],
        "nondesktop_nofire_tier1_2020": [s2d('2017-11-10'), s2d('2018-03-11')],
    }
    temp = data.copy()
    if product in startDates:
        startDate = startDates[product]  # noqa: F841
        temp = temp.query("ds >= @startDate")
    if product in anomalyDates:
        anomalyStartDate = anomalyDates[product][0]  # noqa: F841
        anomalyEndDate = anomalyDates[product][1]  # noqa: F841
        temp = temp.query("(ds < @anomalyStartDate) | (ds > @anomalyEndDate)")
    return temp



import itertools as it
import matplotlib.pyplot as plt


########
# Plot #
########
def plot_tree_aggs(y, levels, no_legend=True, widths=[1, 3], ylabel=None):
    """
    @levels: e.g.: [2, 3, 6, 7, 8] will plot mid point (like median or mean)
    at 6, a thin horizontal bar from 2-8, and a thicker bar from 3-7.
    `y` indicates the height at which the horizontal bar is plotted.
    """
    label = "_nolegend_" if no_legend else None
    v_lo, lo, mid, hi, v_hi = levels
    plt.plot(mid, y, ".k", markersize=10, label=label)

    lines = [(v_lo, v_hi), (lo, hi)]
    for i, (x_lo, x_hi), width in zip(it.count(), lines[::-1], widths[::-1]):
        plt.hlines(y, x_lo, x_hi, linewidth=width, label=label)
    return y, ylabel


def plot_trees_aggs(rel_up, cs=None):
    cs = list(rel_up) if cs is None else cs
    yts = []
    for i, col in ((-i, col) for i, col in zip(it.count(), cs)):
        qs = rel_up[col].mul(100).values[:-1]
        plot_tree_aggs(i, qs, widths=[1, 3], no_legend=i)
        yts.append((i, col))
    plt.yticks(*zip(*yts))
    plt.legend(["mean", "95% CI", "99% CI"])
    return yts, i


def plot_trees_aggs_rel_uplift(rel_up, cs=None):
    yts, i = plot_trees_aggs(rel_up, cs=cs)
    plt.vlines(0, i, [0], "k", linestyle="-.")
    plt.xlabel("% uplift")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from pyspark.sql import functions as F
from pyspark.sql.functions import col, lit
from pyspark.sql import Window
from pyspark.sql.types import IntegerType

from dscontrib.jmccrosky.gudnightly.utils import jackknifeCountCI, jackknifeMeanCI


def metricDAU(data, needed_dimension_variables, feature_col, sampling_multiplier):
    data = data.select(
        ["date", feature_col, "bucket"] + needed_dimension_variables
    ).groupBy(
        ["date", "bucket"] + needed_dimension_variables
    ).agg(
        (F.sum(feature_col) * sampling_multiplier).alias(feature_col)
    )

    data = data.unionByName(
        data.groupBy(
            ["date"] + needed_dimension_variables
        ).agg(
            F.sum(feature_col).alias(feature_col)
        ).withColumn(
            'bucket', F.lit("ALL")
        )
    )
    return data


def metricMAU(data, needed_dimension_variables, feature_col, sampling_multiplier):
    days = 28
    all_user_days = data.select("id").distinct().crossJoin(
        data.select("date").distinct()
    )

    data = data.select(
        ["id", "date", feature_col, "bucket"] + needed_dimension_variables
    ).filter(col(feature_col) == 1)

    data = data.alias("data")
    all_user_days = all_user_days.alias("all_user_days")

    # Augment activity table to include non-active days
    data = data.join(
        all_user_days,
        ['id', 'date'],
        'outer'
    ).withColumn(
        "n_", F.coalesce("data." + feature_col, lit(0))
    ).drop(
        feature_col
    ).withColumnRenamed(
        "n_", feature_col
    )

    if days > 1:
        # Aggregate active days over time for each profile-day
        windowSpec = Window.partitionBy(
            [data.id]
        ).orderBy(
            data.date
        ).rowsBetween(
            1-days, 0
        )

        data = data.withColumn(
            "n_", F.max(data[feature_col]).over(windowSpec) * sampling_multiplier
        ).drop(
            feature_col
        ).withColumnRenamed(
            "n_", feature_col
        )
    for v in needed_dimension_variables:
        data = data.withColumn(v, F.last(v, True).over(windowSpec))

    # Reduce to sum of active days per date
    data = data.filter(
        col(feature_col) > 0
    ).groupBy(
        ["date", "bucket"] + needed_dimension_variables
    ).sum(
        feature_col
    ).withColumnRenamed(
        'sum({})'.format(feature_col),
        feature_col
    )

    # Add all-bucket rows
    data = data.unionByName(
        data.groupBy(
            ["date"] + needed_dimension_variables
        ).agg(
            F.sum(feature_col).alias(feature_col)
        ).withColumn(
            'bucket', lit("ALL")
        )
    )
    return data


# This function won't work with gudnightly currently and is a utility for
# other analysis.
def metricActiveHoursPerWeekPerProfileDay(
    data, needed_dimension_variables, feature_col, sampling_multiplier, days=7,
    include_day_of_week=False
):
    all_user_days = data.select("id").distinct().crossJoin(
        data.select("date").distinct()
    )
    intermediate_table1 = data.filter(
        col(feature_col) > 0
    ).select(
        ["id", "date", feature_col]
    ).distinct(
    )

    intermediate_table1 = intermediate_table1.alias("intermediate_table")
    all_user_days = all_user_days.alias("all_user_days")

    # Augment activity table to include non-active days
    intermediate_table2 = intermediate_table1.join(
        all_user_days,
        ['id', 'date'],
        'outer'
    ).withColumn(
        "n_", F.coalesce("intermediate_table." + feature_col, lit(0))
    ).drop(
        feature_col
    ).withColumnRenamed(
        "n_", feature_col
    )

    if include_day_of_week:
        intermediate_table2 = intermediate_table2.withColumn(
            "weekday_hours_sum",
            F.when(
                F.date_format('date', 'u').cast(IntegerType()) <= 5,
                col("active_hours_sum")
            ).otherwise(0)
        ).withColumn(
            "weekend_hours_sum",
            F.when(
                F.date_format('date', 'u').cast(IntegerType()) >= 6,
                col("active_hours_sum")
            ).otherwise(0)
        )

    # Calculate active hours for each profile-day
    windowSpec = Window.partitionBy(
        [intermediate_table2.id]
    ).orderBy(
        intermediate_table2.date
    ).rowsBetween(
        1 - days, 0
    )

    active_hours_table = intermediate_table2.withColumn(
        "hours", F.sum(intermediate_table2["active_hours_sum"]).over(windowSpec)
    )
    if include_day_of_week:
        active_hours_table = active_hours_table.withColumn(
            "weekday_hours",
            F.sum(intermediate_table2["weekday_hours_sum"]).over(windowSpec)
        ).withColumn(
            "weekend_hours",
            F.sum(intermediate_table2["weekend_hours_sum"]).over(windowSpec)
        )
    return active_hours_table


def metricSumDimensionOverWeekPerProfileDay(
    data, needed_dimension_variables, feature_col, sampling_multiplier, days=7,
    include_day_of_week=False
):
    all_user_days = data.select("id").distinct().crossJoin(
        data.select("date").distinct()
    )
    intermediate_table1 = data.filter(
        col(feature_col) > 0
    ).select(
        ["id", "date", feature_col]
    ).distinct(
    )

    intermediate_table1 = intermediate_table1.alias("intermediate_table")
    all_user_days = all_user_days.alias("all_user_days")

    # Augment activity table to include non-active days
    intermediate_table2 = intermediate_table1.join(
        all_user_days,
        ['id', 'date'],
        'outer'
    ).withColumn(
        "n_", F.coalesce("intermediate_table." + feature_col, lit(0))
    ).drop(
        feature_col
    ).withColumnRenamed(
        "n_", feature_col
    )

    if include_day_of_week:
        intermediate_table2 = intermediate_table2.withColumn(
            "weekday_" + feature_col,
            F.when(
                F.date_format('date', 'u').cast(IntegerType()) <= 5,
                col(feature_col)
            ).otherwise(0)
        ).withColumn(
            "weekend_" + feature_col,
            F.when(
                F.date_format('date', 'u').cast(IntegerType()) >= 6,
                col(feature_col)
            ).otherwise(0)
        )

    # Calculate active hours for each profile-day
    windowSpec = Window.partitionBy(
        [intermediate_table2.id]
    ).orderBy(
        intermediate_table2.date
    ).rowsBetween(
        1 - days, 0
    )

    active_hours_table = intermediate_table2.withColumn(
        "_temp", F.sum(intermediate_table2[feature_col]).over(windowSpec)
    )
    if include_day_of_week:
        active_hours_table = active_hours_table.withColumn(
            "_temp_weekday",
            F.sum(intermediate_table2["weekday_" + feature_col]).over(windowSpec)
        ).withColumn(
            "_temp_weekend",
            F.sum(intermediate_table2["weekend_" + feature_col]).over(windowSpec)
        )
    return active_hours_table.drop(
        feature_col,
        "weekday_" + feature_col,
        "weekend_" + feature_col,
    ).withColumnRenamed(
        "_temp",
        feature_col
    ).withColumnRenamed(
        "_temp_weekday",
        "weekday_" + feature_col
    ).withColumnRenamed(
        "_temp_weekend",
        "weekend_" + feature_col
    )


def metricDaysPerWeekPerProfileDay(
    data, needed_dimension_variables, feature_col, sampling_multiplier, days=7,
    include_day_of_week=False
):
    all_user_days = data.select("id").distinct().crossJoin(
        data.select("date").distinct()
    )

    data = data.filter(
        col(feature_col) > 0
    ).select(
        ["id", "date", "bucket", feature_col] + needed_dimension_variables
    ).distinct(
    )

    data = data.alias("intermediate_table")
    all_user_days = all_user_days.alias("all_user_days")

    # Augment activity table to include non-active days
    intermediate_table2 = data.join(
        all_user_days,
        ['id', 'date'],
        'outer'
    ).withColumn(
        "n_", F.coalesce("intermediate_table." + feature_col, lit(0))
    ).drop(
        feature_col
    ).withColumnRenamed(
        "n_", feature_col
    )

    if include_day_of_week:
        intermediate_table2 = intermediate_table2.withColumn(
            feature_col + "_weekend",
            F.when(
                F.date_format('date', 'u').cast(IntegerType()) >= 6,
                col(feature_col)
            ).otherwise(
                0
            )
        ).withColumn(
            feature_col + "_weekday",
            F.when(
                F.date_format('date', 'u').cast(IntegerType()) <= 5,
                col(feature_col)
            ).otherwise(
                0
            )
        )

    # Calculate active days per week for each profile-day
    windowSpec = Window.partitionBy(
        [intermediate_table2.id]
    ).orderBy(
        intermediate_table2.date
    ).rowsBetween(
        1 - days, 0
    )

    intermediate_table3 = intermediate_table2.withColumn(
        "n_", F.sum(intermediate_table2[feature_col]).over(windowSpec)
    ).drop(
        feature_col
    ).withColumnRenamed(
        "n_", feature_col
    )
    if include_day_of_week:
        intermediate_table3 = intermediate_table3.withColumn(
            "n_", F.sum(intermediate_table2[feature_col + "_weekend"]).over(windowSpec)
        ).drop(
            feature_col + "_weekend"
        ).withColumnRenamed(
            "n_", feature_col + "_weekend"
        )

        intermediate_table3 = intermediate_table3.withColumn(
            "n_", F.sum(intermediate_table2[feature_col + "_weekday"]).over(windowSpec)
        ).drop(
            feature_col + "_weekday"
        ).withColumnRenamed(
            "n_", feature_col + "_weekday"
        )

    for v in needed_dimension_variables:
        intermediate_table3 = intermediate_table3.withColumn(
            v, F.last(v, True).over(windowSpec)
        )
    return intermediate_table3


def metricDaysPerWeek(
    data, needed_dimension_variables, feature_col, sampling_multiplier
):
    intermediate_table3 = metricDaysPerWeekPerProfileDay(
        data, needed_dimension_variables, feature_col, sampling_multiplier
    )

    # Reduce to mean active days per week per date
    intermediate_table4 = intermediate_table3.filter(
        col(feature_col) > 0
    ).groupBy(
        ["date", "bucket"] + needed_dimension_variables
    ).mean(
        feature_col
    ).withColumnRenamed(
        'avg({})'.format(feature_col),
        feature_col
    )
    intermediate_table4_allbucket = intermediate_table3.filter(
        col(feature_col) > 0
    ).groupBy(
        ["date"] + needed_dimension_variables
    ).mean(
        feature_col
    ).withColumnRenamed(
        'avg({})'.format(feature_col),
        feature_col
    ).withColumn(
        'bucket', lit("ALL")
    )

    # Add all-bucket rows
    joined_intermediate = intermediate_table4.unionByName(
        intermediate_table4_allbucket
    )
    return joined_intermediate


def metricRetention(
    data, needed_dimension_variables, feature_col, sampling_multiplier,
    activated=False
):
    activity_data = data.filter(
        col(feature_col) > 0
    ).select(
        ["id", "date", feature_col]
    ).distinct()

    pcd_table = data.select(
        ["date", "id", "bucket"] + needed_dimension_variables
    )
    windowSpec = Window.partitionBy(
        [pcd_table.id] + needed_dimension_variables
    ).orderBy(
        pcd_table.date
    ).rowsBetween(
        0, 13
    )
    for v in needed_dimension_variables:
        pcd_table = pcd_table.withColumn(v, F.last(v, True).over(windowSpec))
    pcd_table = pcd_table.filter(
        col("new_profile") == 1
    )

    if activated:
        pcd_table = pcd_table.alias("pcd_t").join(
            activity_data.alias("i_t"),
            (col('pcd_t.id') == col('i_t.id')) &
            (col('i_t.date') >= F.date_add(col('pcd_t.date'), 1)) &
            (col('i_t.date') <= F.date_add(col('pcd_t.date'), 6)),
            "inner"
        ).filter(
            col("i_t." + feature_col) > 0
        ).dropDuplicates(
            ['id']
        ).select(
            [
                col('pcd_t.{}'.format(c))
                for c in ['id', 'bucket', "date"] + needed_dimension_variables
            ]
        )

    intermediate_table3 = pcd_table.alias("pcd_t").join(
        activity_data.alias("i_t"),
        (col('pcd_t.id') == col('i_t.id')) &
        (col('i_t.date') >= F.date_add(col('pcd_t.date'), 7)) &
        (col('i_t.date') <= F.date_add(col('pcd_t.date'), 13)),
        "outer"
    ).select(
        [
            'pcd_t.{}'.format(c)
            for c in ['id', 'date', 'bucket'] +
            needed_dimension_variables
        ] + [feature_col]
    ).fillna(
        0, [feature_col]
    ).groupBy(
        [
            'pcd_t.{}'.format(c)
            for c in ['id', 'date', 'bucket'] +
            needed_dimension_variables
        ],
    ).agg(
        F.max(col(feature_col))
    ).drop(
        feature_col
    ).withColumnRenamed(
        "MAX({})".format(feature_col), feature_col
    ).select(
        [
            col("pcd_t.{}".format(c)).alias(c)
            for c in ['id', 'bucket', 'date'] +
            needed_dimension_variables
        ] + [feature_col]
    )

    intermediate_table4 = intermediate_table3.groupBy(
        ["date", "bucket"] + needed_dimension_variables
    ).mean(
        feature_col
    ).withColumnRenamed(
        'avg({})'.format(feature_col), feature_col
    )
    intermediate_table4_allbucket = intermediate_table3.groupBy(
        ["date"] + needed_dimension_variables
    ).mean(
        feature_col
    ).withColumnRenamed(
        'avg({})'.format(feature_col), feature_col
    ).withColumn(
        'bucket', lit("ALL")
    )

    joined_intermediate = intermediate_table4.unionByName(
        intermediate_table4_allbucket
    )
    return joined_intermediate


def metricActivatedRetention(
    data, needed_dimension_variables, feature_col, sampling_multiplier
):
    return metricRetention(
        data, needed_dimension_variables, feature_col, sampling_multiplier, True
    )


metricFunctions = {
    "DAU": metricDAU,
    "MAU": metricMAU,
    # "MAU31": metricMAU31,
    # "MAU27": metricMAU27,
    "Days Per Week": metricDaysPerWeek,
    "Week 1 Retention": metricRetention,
    "Week 1 Activated Retention": metricActivatedRetention,
    # "Week 1 Retention (excluding single-day profiles)": metricRealishRetention,
}

metricAggregations = {
    "DAU": lambda x: x.sum(),
    "MAU": lambda x: x.sum(),
    # "MAU31": lambda x: x.sum(),
    # "MAU27": lambda x: x.sum(),
    "Days Per Week": lambda x: x.sum() if len(x) == 1 else 0,
    "Week 1 Retention": lambda x: x.sum() if len(x) == 1 else 0,
    "Week 1 Activated Retention": lambda x: x.sum() if len(x) == 1 else 0,
    # "Week 1 Retention (excluding single-day profiles)":
    #      lambda x: x.sum() if len(x)==1 else 0,
}

metricDaysNeededPre = {
    "DAU": 0,
    "MAU": 27,
    # "MAU31": 30,
    # "MAU27": 26,
    "Days Per Week": 6,
    "Week 1 Retention": 0,
    "Week 1 Activated Retention": 0,
    # "Week 1 Retention (excluding single-day profiles)": 0,
}

metricDaysNeededPost = {
    "DAU": 0,
    "MAU": 0,
    # "MAU31": 0,
    # "MAU27": 0,
    "Days Per Week": 0,
    "Week 1 Retention": 13,
    "Week 1 Activated Retention": 13,
    # "Week 1 Retention (excluding single-day profiles)": 13,
}

metricCIs = {
    "DAU": jackknifeCountCI,
    # "DAU-alt": poissonCI,
    "MAU": jackknifeCountCI,
    # "MAU31": jackknifeCountCI,
    # "MAU27": jackknifeCountCI,
    # "Retention-debug": jackknifeCountCI,
    "Days Per Week": jackknifeMeanCI,
    "Week 1 Retention": jackknifeMeanCI,
    "Week 1 Activated Retention": jackknifeMeanCI,
    # "Week 1 Retention-alt": binomialCI,
    # "Week 1 Retention (excluding single-day profiles)": jackknifeMeanCI,
    # "Retention-filter": jackknifeMeanCI,
    # "WAU": jackknifeCountCI,
}

metricHTs = {
    # "DAU": permutationTestCount,
    # "Days Per Week": permutationTestMean,
    # "MAU": permutationTestCount,
}



from functools import wraps
from pathlib import Path
import shutil

import joblib
import pandas as pd


def parquet_cache(mem: joblib.Memory, engine="auto"):
    """
    mem = joblib.Memory(cachedir="cache", verbose=0)

    @parquet_cache(mem)
    def basic_df(a):
        time.sleep(1)
        return DataFrame(dict(a=[a] * 10))

    basic_df.clear() to empty
    """

    def mk_deco(fn):
        job_lib_func = mem.cache(fn)
        path = Path(
            job_lib_func.store_backend.location
        ) / joblib.memory._build_func_identifier(fn)

        def clear():
            shutil.rmtree(path)

        @wraps(fn)
        def deco(*a, **k):
            arg_hash = job_lib_func._get_argument_hash(*a, **k)
            loc = path / f"{arg_hash}.pq"
            if loc.exists():
                return pd.read_parquet(loc, engine=engine)
            df = fn(*a, **k)
            df.to_parquet(loc, engine=engine)
            return df

        deco.path = path
        deco.clear = clear
        return deco

    return mk_deco



# date utils

# date_plus_N(date, N)
# date_to_string(date, format='%Y%m%d')
# string_to_date(date, format='%Y%m%d')
# date_range(start_dt, N)

# spark utils

# read_parquet_from_s3(path)
# write_parquet_to_s3(df, path, mode='overwrite')
# read_main_summary()


# date util: imports
import datetime
from datetime import timedelta

# spark util: imports
from pyspark.sql import SparkSession

# --------------- date utils ---------------


def date_plus_N(date, N):
    """
    takes datetime.date
    return another datetime.date N days away
    """
    return date + timedelta(days=N)


def date_to_string(date, format='%Y%m%d'):
    """
    takes datetime.date
    returns it back in string form
    """
    return date.strftime(format)


def string_to_date(date, format='%Y%m%d'):
    """
    takes string
    returns datetime.date
    """
    return datetime.datetime.strptime(date, format).date()


def date_range(start_dt, N):
    """
    N can be int or a datetime.date
    returns an array of datetime.date
    """
    if type(N) != int:
        N = (N - start_dt).days
    return [date_plus_N(start_dt, n) for n in range(N + 1)]


# --------------- spark utils ---------------


spark = SparkSession.builder.getOrCreate()


def read_parquet_from_s3(path):
    """
    read a df from s3 source
        returns df
    """
    return spark.read.parquet(path)


def write_parquet_to_s3(df, path, mode='overwrite'):
    """
    write a df to s3 source
    """
    df.write.mode(mode).parquet(path)
    print('data saved to: %s' % path)


def read_main_summary():
    """
    load ms directly from s3
        returns df
    """
    ms_path = 's3://telemetry-parquet/main_summary/v4/'
    return spark.read.option("mergeSchema", "true").parquet(ms_path)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


import pandas as pd


_queries = {
    "light_funnel_sampled_dau_city": '''
        WITH top_cities_t AS (
          SELECT
            COUNT(client_id) AS dau,
            CONCAT(
              country, ":",
              geo_subdivision1, ":",
              geo_subdivision2, ":",
              city
            ) AS geo
          FROM `moz-fx-data-shared-prod.telemetry.clients_daily`
          WHERE
            submission_date = "2020-03-01"
            AND city != "??"
            AND sample_id=67
          GROUP BY
            country,
            geo_subdivision1,
            geo_subdivision2,
            city
          ORDER BY dau DESC
          LIMIT 1000
        ),
        geo_t AS (
          SELECT
            client_id,
            geo
          FROM (
            SELECT
              client_id,
              CONCAT(
                country, ":",
                geo_subdivision1, ":",
                geo_subdivision2, ":",
                cd_t.city
              ) AS geo,
              ROW_NUMBER() OVER(
                PARTITION BY client_id ORDER BY submission_date DESC
              ) AS rn
            FROM `moz-fx-data-shared-prod.telemetry.clients_daily` AS cd_t
            INNER JOIN (SELECT geo FROM top_cities_t) AS top_cities_t
            ON CONCAT(
                country, ":",
                geo_subdivision1, ":",
                geo_subdivision2, ":",
                city
              ) = top_cities_t.geo
            WHERE
              submission_date > "1900-01-01"
              AND sample_id=67
          )
          WHERE rn = 1
        )
        SELECT
          submission_date AS date,
          COUNT(client_id) * 100 AS value,
          geo_t.geo AS geo
        FROM (
            SELECT
              client_id,
              submission_date,
            FROM `moz-fx-data-shared-prod.telemetry.clients_daily`
            WHERE
              submission_date > "1900-01-01"
              AND sample_id=67
              AND attribution.content IS NOT NULL
          )
        INNER JOIN geo_t
        USING(client_id)
        GROUP BY geo, submission_date
        '''
}


def get_raw_data(bq_client, bq_storage_client, metric):
    return bq_client.query(
        _queries[metric]
    ).result().to_dataframe(bqstorage_client=bq_storage_client)


def prepare_data(data, training_start, training_end):
    clean_data = {}
    clean_training_data = {}
    # Suppress any geoXdate with less than 5000 profiles as per minimum
    # aggregation standards for the policy this data will be released under.
    data = data[data.value >= 5000]
    for c in data.geo.unique():
        if (len(data.query("geo==@c")) < 100):
            continue
        clean_data[c] = data.query("geo==@c").rename(
            columns={"date": "ds", "value": "y"}
        ).sort_values("ds")
        clean_data[c]['ds'] = pd.to_datetime(clean_data[c]['ds']).dt.date
        clean_data[c] = clean_data[c].set_index('ds')
        clean_training_data[c] = clean_data[c][
            training_start:training_end
        ].reset_index()
        clean_data[c] = clean_data[c].reset_index()
    return (clean_data, clean_training_data)



import datetime as dt
import itertools as it
import re
from typing import List, Optional, Tuple, Union

import pandas as pd  # type: ignore
from pandas import DataFrame
from requests import post

uri = "https://buildhub.moz.tools/api/search"


def get_major(s):
    "str -> int"
    return int(s.split(".")[0])


def pull_build_id_docs(min_build_day="20180701", channel="beta", raw_json=False):
    """
    Note, we're only taking win64, en-US, assuming that other build_ids will
    just be duplicates.
    """
    query = {
        "aggs": {
            "buildid": {
                "terms": {
                    "field": "build.id",
                    "size": 100000,
                    "order": {"_term": "desc"},
                },
                # "aggs": {"version": {"terms": {"field": "target.version"}}},
                "aggs": {
                    "version": {"terms": {"field": "target.version"}},
                    "pub_date": {"terms": {"field": "download.date"}},
                    "buildid": {"terms": {"field": "build.id"}},
                },
            }
        },
        "query": {
            "bool": {
                "filter": [
                    {"term": {"target.platform": "win64"}},
                    {"term": {"target.locale": "en-US"}},
                    {"term": {"target.channel": channel}},
                    {"term": {"source.product": "firefox"}},
                    {"range": {"build.id": {"gte": min_build_day}}},
                ]
            }
        },
        "size": 0,
    }
    resp = post(uri, json=query)
    json = resp.json()
    if raw_json:
        return json
    docs = json["aggregations"]["buildid"]["buckets"]
    return docs


def extract_triplets(
    doc,
    major_version: Union[int, None, Tuple[int], List[int]] = None,
    keep_rc=False,
    keep_release=False,
    agg=min,
):
    """
    @major_version: int, Tuple[int] or (callable: str -> bool)
    [doc] = aggregations.buildid.buckets ->
        doc.version.buckets[].key
    From json results, return tuple of `buildids`, `pub_dates`, `versions`

    Some build_id's go to multiple versions (usually rc's or a major version)
    Some versions go to multiple build_id's.
    """

    def collect_results(field):
        return [res["key"] for res in doc[field]["buckets"]]

    def major_version_filt(v):
        if major_version is None:
            return True
        elif isinstance(major_version, int):
            return get_major(v) == major_version
        elif isinstance(major_version, (tuple, list)):
            return get_major(v) in major_version
        else:
            return major_version(v)

    buildids = collect_results("buildid")
    pub_dates = collect_results("pub_date")
    versions = [
        v
        for v in collect_results("version")
        if version_filter(v, keep_rc=keep_rc, keep_release=keep_release)
        and major_version_filt(v)
    ]
    if not versions:
        return None
    return agg(versions), agg(buildids), agg(pub_dates)


def version2build_ids(
    docs, major_version=None, keep_rc=False, keep_release=False, as_df=False
):
    version_build_ids = [
        extract_triplets(
            doc, major_version=major_version, keep_rc=keep_rc, keep_release=keep_release
        )
        for doc in docs
    ]
    version_build_ids = filter(None, version_build_ids)
    dct = {
        version: [
            (_version, build_id, pub_date) for _version, build_id, pub_date in triplets
        ]
        for version, triplets in it.groupby(version_build_ids, lambda x: x[0])
    }
    if as_df:
        return (
            pd.DataFrame(
                [
                    trip
                    for trips in version2build_ids(
                        docs, keep_release=True, keep_rc=False
                    ).values()
                    for trip in trips
                ],
                columns=["dvers", "build_id", "pub_date"],
            )
            .assign(pub_date=lambda x: pd.to_datetime(x.pub_date, unit="ms"))
            .sort_values(["pub_date"], ascending=False)
            .reset_index(drop=1)
        )
    return dct


def rc_major_version(disp_vers: str) -> Optional[int]:
    rc_re = re.compile(r"(?P<major>\d\d+)(\.\d)+$")
    m = rc_re.match(disp_vers)
    return int(m.group("major")) if m else None


def pull_beta_rc_builds(beta_major_versions, min_build_day="20180701"):
    """
    """

    def combine_bid_rc(rc_major, build_id):
        if rc_major == rc_major:
            return f"{int(rc_major)}rc{build_id}"
        return rc_major

    docs = pull_build_id_docs(
        min_build_day=min_build_day, channel="beta", raw_json=False
    )
    bhdf = (
        version2build_ids(docs, keep_release=True, keep_rc=False, as_df=True)
        .assign(rc_major=lambda x: x.dvers.map(rc_major_version))
        .assign(is_rc=lambda x: x.rc_major.notnull())
        .assign(
            dvers2=lambda x: [
                combine_bid_rc(rc, bid) for rc, bid in zip(x.rc_major, x.build_id)
            ]
        )
        .assign(
            dvers=lambda x: x[["dvers", "dvers2"]].fillna(axis=1, method="ffill").dvers2
        )
    )

    return bhdf


def version2build_id_str(docs, major_version=None, keep_rc=False, keep_release=False):
    """
    Returns mapping of display version to 'sql' list of build_ids.
    E.g., {'70.0b3': "'20190902191027', '20190902160204', '20190902120346'"}
    """
    triplet_dct = version2build_ids(
        docs, major_version=major_version, keep_rc=keep_rc, keep_release=keep_release
    )
    return {
        dvers: ", ".join(["'{}'".format(bid) for _dvers, bid, pub_date in trips])
        for dvers, trips in triplet_dct.items()
    }


def version2df(docs, major_version=None, keep_rc=False, keep_release=False):
    """
    Given docs from ---
    return DataFrame with columns `disp_vers`, `build_id`, `pub_date`.
    """
    triples = version2build_ids(
        docs, major_version=major_version, keep_rc=keep_rc, keep_release=keep_release
    )
    df = (
        DataFrame(
            [trip for trips in triples.values() for trip in trips],
            columns=["disp_vers", "build_id", "pub_date"],
        )
        .assign(
            pub_date=lambda x: (x.pub_date // 1e3)
            .map(dt.datetime.utcfromtimestamp)
            .pipe(pd.to_datetime)
        )
        .sort_values(["pub_date"], ascending=True)
        .reset_index(drop=1)
    )

    return df


def version_filter(x, keep_rc=False, keep_release=False):
    if not keep_rc and "rc" in x:
        return False
    if not keep_release and ("b" not in x):
        return False
    return True


def test_version_filter():
    rc_vers = "65.0b6rc"
    assert version_filter(rc_vers, keep_rc=False, keep_release=False) is False
    assert version_filter(rc_vers, keep_rc=True, keep_release=False)

    rls_vers = "65.0"
    assert version_filter(rls_vers, keep_rc=False, keep_release=True)
    assert version_filter(rls_vers, keep_rc=False, keep_release=False) is False

    beta_vers = "65.0b6"
    assert version_filter(beta_vers, keep_rc=False, keep_release=True)
    assert version_filter(beta_vers, keep_rc=False, keep_release=False)


def months_ago(months=12):
    return (dt.date.today() - dt.timedelta(days=30 * months)).strftime("%Y%m%d")


def pull_channel(
    channel, major_version: Union[int, None, Tuple[int], List[int]], min_build_day=None
):
    min_build_day = min_build_day or months_ago(12)
    result_docs = pull_build_id_docs(min_build_day=min_build_day, channel=channel)
    return version2build_ids(
        result_docs,
        major_version=major_version,
        keep_rc=False,
        keep_release=True,
        as_df=True,
    ).assign(mvers=lambda x: x.dvers.str.split(".").str[0].astype(int))


def main(vers=67):
    # result_docs = pull_build_id_docs(min_build_day="20180701")
    result_docs = pull_build_id_docs(min_build_day=months_ago(12), channel="beta")
    res = version2build_ids(
        result_docs, major_version=67, keep_rc=False, keep_release=False
    )
    print(res)


def main_release(vers=67):
    # result_docs = pull_build_id_docs(min_build_day="20180701")
    result_docs = pull_build_id_docs(min_build_day=months_ago(12), channel="release")
    res = version2build_ids(
        result_docs, major_version=67, keep_rc=False, keep_release=True
    )
    print(res)


if __name__ == "__main__":
    main()

    print("\n\nrelease")
    main_release()



import datetime as dt

import pandas as pd
import numpy as np

"""
* clip

"""

# Help with dates
SUB_FMT = "%Y%m%d"
SUB_FMT_DASH = "%Y-%m-%d"
SUB_FMT_SPK = "yyyyMMdd"


class Date:
    def __init__(self, date_str):
        self.date_str = date_str
        self.d = d = pd.to_datetime(date_str)
        self.s = d.strftime(SUB_FMT)
        self.sd = d.strftime(SUB_FMT_DASH)

    def __repr__(self):
        return "Date({})".format(self.sd)

    def add_days(self, n_days):
        return Date(add_days(n_days, self.d))

    def __eq__(self, o):
        return self.d.date() == o.d.date()

    def __sub__(self, o):
        return (self.d - o.d).days

    @staticmethod
    def today():
        return Date(dt.date.today())


def add_days(n, date):
    ni = int(n)
    if type(ni) != type(n):
        print("Warning, type converted")
    ret = date + dt.timedelta(days=ni)
    return ret


def s3(date):
    return date.strftime(SUB_FMT)


def s3_dash(date):
    return date.strftime(SUB_FMT_DASH)


# Stats
def clip_srs(s, clip_percentile=99.9, ignore_nulls=True, vb=False):
    if ignore_nulls:
        val = np.percentile(s[s == s], clip_percentile)
    else:
        val = np.percentile(s, clip_percentile)
    if vb:
        print("{} -> {}".format(s.max(), val))
    sc = np.clip(s, None, val)
    return sc



#!/usr/bin/env python
# Helper for setuptools-scm.

import datetime as dt
import subprocess

ref = subprocess.check_output("git describe --dirty --all --long --first-parent --match master".split()).strip().decode("utf-8")
last_commit = subprocess.check_output('git show -s --format=format:%ct master'.split()).strip().decode("ascii")

version_string = dt.datetime.utcfromtimestamp(int(last_commit)).strftime("%Y%m%d%H%M%S")
print(ref.replace("heads/master", version_string))



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import numpy as np
import pandas as pd
from pyspark.sql.functions import col
import plotly.graph_objs as go
from plotly.offline import plot

from dscontrib.jmccrosky.gudnightly.usage_criteria import usage_criteria
from dscontrib.jmccrosky.gudnightly.metrics import metricFunctions, metricAggregations
from dscontrib.jmccrosky.gudnightly.metrics import metricDaysNeededPre
from dscontrib.jmccrosky.gudnightly.metrics import metricCIs, metricDaysNeededPost
from dscontrib.jmccrosky.gudnightly.utils import calculateDateWindow, doSmoothing
from dscontrib.jmccrosky.gudnightly.utils import getPandasDimensionQuery, dimensionName
from dscontrib.jmccrosky.gudnightly.utils import jackknifeMeanCI, longDimensionName


def MetricPlot(
    feature_data,
    plot_start_date, plot_end_date,
    criterium, dimensions, metric,
    jackknife_buckets,
    sampling_multiplier,
    smoothing=1,
    extra_filter="TRUE",
    comparison_mode="None",
    transformations=[],
    force_width=None,
    force_height=None,
    x_min=None,
    x_max=None,
    y_min=None,
    y_max=None,
    suppress_ci=False,
    debug="No"
):
    feature_col = usage_criteria[criterium]
    needed_dimension_variables = list(set().union(*(d.keys() for d in dimensions)))
    buckets_list = ["{:d}".format(i) for i in range(jackknife_buckets)]
    date_window = calculateDateWindow(
        plot_start_date, plot_end_date, smoothing, comparison_mode,
        metricDaysNeededPre[metric], metricDaysNeededPost[metric]
    )
    if len(date_window) == 1:
        x = date_window[0]
        intermediate_table = feature_data.filter(feature_data.date.between(x[0], x[1]))
    elif len(date_window) == 2:
        x = date_window[0]
        y = date_window[1]
        intermediate_table = feature_data.filter(
            feature_data.date.between(x[0], x[1]) |
            feature_data.date.between(y[0], y[1])
        )
    else:
        print("need to find a nice way to handle this")
        return
        intermediate_table = intermediate_table.filter(
            extra_filter
        )

    data = metricFunctions[metric](
        intermediate_table, needed_dimension_variables, feature_col, sampling_multiplier
    )

    if smoothing > 1:
        data = doSmoothing(data, feature_col, needed_dimension_variables, smoothing - 1)

    data = data.filter(
        col('date').between(pd.to_datetime("19000101"), pd.to_datetime("21000101"))
    ).toPandas()
    data['date'] = pd.to_datetime(data.date)
    if debug == "data":
        return data

    if "Day Of Week" in transformations:
        dates = [
            "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday"
        ]
    else:
        dates = pd.date_range(plot_start_date, plot_end_date)
    plot_data = {
        "date": dates,
    }

    if comparison_mode == "Slices":
        comparison_dimensions = dimensions
        dimensions = [{}]
    for dim in dimensions:
        if "Day Of Week" in transformations:
            data['day_of_week'] = data['date'].dt.day_name()
            date_var = "day_of_week"
        else:
            date_var = "date"
        cur_data = data[
            data["date"].isin(pd.date_range(plot_start_date, plot_end_date))
        ]
        if comparison_mode == "Slices":
            cur_all1 = [metricAggregations[metric](
                cur_data.query(
                    "bucket=='ALL' and {}==@d".format(date_var) +
                    getPandasDimensionQuery(comparison_dimensions[0])
                )[feature_col]
            ) for d in dates]
            cur_buckets1 = [[metricAggregations[metric](
                cur_data.query(
                    "bucket==@i and {}==@d".format(date_var) +
                    getPandasDimensionQuery(comparison_dimensions[0]),
                    local_dict=dict(d=d, i=i)
                )[feature_col]
            ) for i in buckets_list] for d in dates]
            cur_all2 = [metricAggregations[metric](
                cur_data.query(
                    "bucket=='ALL' and {}==@d".format(date_var) +
                    getPandasDimensionQuery(comparison_dimensions[1]),
                    local_dict=dict(d=d)
                )[feature_col]) for d in dates]
            cur_buckets2 = [[metricAggregations[metric](
                cur_data.query(
                    "bucket==@i and {}==@d".format(date_var) +
                    getPandasDimensionQuery(comparison_dimensions[1]),
                    local_dict=dict(d=d, i=i)
                )[feature_col]
            ) for i in buckets_list] for d in dates]
        else:
            cur_all = [metricAggregations[metric](
                cur_data.query(
                    "bucket=='ALL' and {}==@d".format(date_var) +
                    getPandasDimensionQuery(dim)
                )[feature_col]
            ) for d in dates]
            cur_buckets = [[metricAggregations[metric](
                cur_data.query(
                    "bucket==@i and {}==@d".format(date_var) +
                    getPandasDimensionQuery(dim),
                    local_dict=dict(d=d, i=i)
                )[feature_col]
            ) for i in buckets_list] for d in dates]

        if comparison_mode in ["Last Year", "YoY"]:
            old_data = data[data["date"].isin(pd.date_range(
                plot_start_date - pd.DateOffset(years=1),
                plot_end_date - pd.DateOffset(years=1)
            ))]
            if "Day Of Week" in transformations:
                def date_offset(x):
                    return x
            else:
                def date_offset(x):
                    return x - pd.DateOffset(years=1)
            old_all = [metricAggregations[metric](old_data.query(
                "bucket=='ALL' and {}==@d".format(date_var) +
                getPandasDimensionQuery(dim))[feature_col]
            ) for d in (date_offset(dates))]
            old_buckets = [[metricAggregations[metric](old_data.query(
                "bucket==@i and {}==@d".format(date_var) +
                getPandasDimensionQuery(dim),
                local_dict=dict(d=d, i=i))[feature_col]
            ) for i in buckets_list] for d in (date_offset(dates))]

        if "Normalize" in transformations:
            total_all = np.sum(cur_all)
            total_buckets = [np.sum(
                [cur_buckets[d][i] for d in range(len(dates))]
            ) for i in range(len(buckets_list))]
            cur_all = [v / float(total_all) for v in cur_all]
            cur_buckets = [[
                    cur_buckets[d][i] / float(total_buckets[i])
                    for i in range(len(buckets_list))
                ] for d in range(len(dates))
            ]
            if comparison_mode != "None":
                total_all = np.sum(old_all)
                total_buckets = [
                    np.sum([old_buckets[d][i] for d in range(len(dates))])
                    for i in range(len(buckets_list))
                ]
                old_all = [v / float(total_all) for v in old_all]
                old_buckets = [[
                    old_buckets[d][i] / float(total_buckets[i])
                    for i in range(len(buckets_list))
                ] for d in range(len(dates))]
        if comparison_mode == "YoY":
            plot_data.update({
                "value" + dimensionName(dim):
                    [(float(c-o)/o) * 100 for c, o in zip(cur_all, old_all)],
                "ci" + dimensionName(dim):
                    [
                        jackknifeMeanCI([(float(i-j)/j) * 100 for i, j in zip(c, o)])
                        for c, o in zip(cur_buckets, old_buckets)
                    ],
            })
        if comparison_mode == "Slices":
            plot_data.update({
                "value" + dimensionName(dim):
                    [(float(c1)/c2) * 100 for c1, c2 in zip(cur_all1, cur_all2)],
                "ci" + dimensionName(dim):
                    [jackknifeMeanCI(
                        [(float(i)/j) * 100 for i, j in zip(c1, c2)]
                    ) for c1, c2 in zip(cur_buckets1, cur_buckets2)],
            })
        if comparison_mode == "Last Year":
            plot_data.update({
                "value" + dimensionName(dim): cur_all,
                "ci" + dimensionName(dim): [metricCIs[metric](i) for i in cur_buckets],
                "value" + dimensionName(dim) + " last year": old_all,
                "ci" + dimensionName(dim) + " last year":
                    [metricCIs[metric](i) for i in old_buckets],
            })
        if comparison_mode == "None":
            plot_data.update({
                "value" + dimensionName(dim): cur_all,
                "ci" + dimensionName(dim):
                    [metricCIs[metric](i) for i in cur_buckets],
            })

    plot_data = pd.DataFrame(plot_data)

    if debug == "plot_data":
        return plot_data

    plotly_data = [
        go.Scatter(
            name=(
                longDimensionName(dim) if comparison_mode != "Slices"
                else "{} as % of {}".format(
                    longDimensionName(comparison_dimensions[0]),
                    longDimensionName(comparison_dimensions[1])
                )
            ),
            x=plot_data.date,
            y=plot_data["value" + dimensionName(dim)],
            error_y=(dict(
                type='data',
                symmetric=False,
                array=[ci[0] for ci in plot_data["ci" + dimensionName(dim)]],
                arrayminus=[c[1] for c in plot_data["ci" + dimensionName(dim)]],
            ) if not suppress_ci else None)
        ) for dim in dimensions] + ([
            go.Scatter(
                name=longDimensionName(dim) + " last year",
                x=plot_data.date,
                y=plot_data["value" + dimensionName(dim) + " last year"],
                error_y=(dict(
                    type='data',
                    symmetric=False,
                    array=[
                        ci[0] for ci in
                        plot_data["ci" + dimensionName(dim) + " last year"]
                    ],
                    arrayminus=[
                        ci[1] for ci in
                        plot_data["ci" + dimensionName(dim) + " last year"]
                    ],
                ) if not suppress_ci else None)
            )
            for dim in dimensions] if comparison_mode == "Last Year" else []
    )
    x_label = "Date"
    ytitle = '{}{}'.format(
        metric, (
            " (Proportion)" if "Normalize" in transformations
            else (" (YoY Percent Change)" if comparison_mode == "YoY" else "")
        )
    )
    if comparison_mode == "Slices":
        ytitle = "Percent of {}".format(metric)
    if (metric in [
        "Week 1 Retention (excluding single-day profiles)",
        "Week 1 Retention"
    ]) or (criterium == "New Profile"):
        x_label = "Profile Creation Date"
    layout = go.Layout(
        showlegend=(True if comparison_mode == "Slices" else None),
        autosize=force_width is None,
        width=force_width,
        height=force_height,
        title=(
            '{} ({}) Over Time'.format(metric, criterium) + (
                " (Restricted to {})".format(dimensionName(dimensions[0]))
                if (len(dimensions) == 1 and len(dimensions[0]) != 0) else ""
            )
        ),
        xaxis=dict(
            title=x_label,
            titlefont=dict(
                family='Courier New, monospace',
                size=18,
                color='#7f7f7f'
            ),
            range=([x_min, x_max] if x_min is not None else None)
        ),
        yaxis=dict(
            title=ytitle,
            titlefont=dict(
                family='Courier New, monospace',
                size=18,
                color='#7f7f7f'
            ),
            range=([y_min, y_max] if y_min is not None else None)
        )
    )
    return plot({"data": plotly_data, "layout": layout}, output_type='div')



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


from dscontrib.jmccrosky.forecast.utils import s2d
from dscontrib.jmccrosky.anomdtct.data import get_raw_data, prepare_data
from dscontrib.jmccrosky.anomdtct.forecast import forecast


def pipeline(bq_client, bq_storage_client):
    raw_data = get_raw_data(
        bq_client,
        bq_storage_client,
        "light_funnel_sampled_dau_city"
    )
    (clean_data, clean_training_data) = prepare_data(
        raw_data, s2d('2016-04-08'), s2d('2019-12-31')
    )
    forecast_data = forecast(clean_training_data, clean_data)
    return forecast_data



# my personal S3 directory
S3_ROOT = 's3://mozilla-metrics/user/shong/'


# activity stream data queries
AS_HEALTH_QUERY = """
      SELECT shield_id, client_id, date, value
      FROM ping_centre_main
      WHERE event = 'AS_ENABLED'
            AND date >= '{START_DT}' AND date <= '{END_DT}'
                  """

AS_SESSION_QUERY = """
    SELECT shield_id, client_id, date, session_id, user_prefs
    FROM assa_sessions_daily_by_client_id
    WHERE date >= '{START_DT}' AND date <= '{END_DT}'
                   """

AS_CLICKS_QUERY = """
  SELECT shield_id, client_id, date, source, page
  FROM assa_events_daily
  WHERE source IN ('TOP_SITES', 'HIGHLIGHTS')
        AND event = 'CLICK'
        AND date >= '{START_DT}' AND date <= '{END_DT}'
                  """

AS_SNIPPETS_QUERY = """
    SELECT shield_id, impression_id AS client_id, date, event, message_id
    FROM assa_router_events_daily
    WHERE source = 'snippets_user_event'
          AND event in ('IMPRESSION', 'BLOCK', 'CLICK_BUTTON', 'DISMISS')
          AND date >= '{START_DT}' AND date <= '{END_DT}'
                    """

# baseline main summary usage columns
MS_USAGE_COLS = ['active_ticks',
                 'search_counts',
                 'scalar_parent_browser_engagement_total_uri_count',
                 'scalar_parent_browser_engagement_tab_open_event_count',
                 'scalar_parent_browser_engagement_window_open_event_count',
                 'scalar_parent_browser_search_with_ads',
                 'scalar_parent_browser_search_ad_clicks',
                 'active_addons']



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


import pandas as pd
from fbprophet import Prophet
from dscontrib.jmccrosky.forecast.utils import s2d

# The only holiday we have identified the need to explicitly model is Chinese
# New Year
chinese_new_year = pd.DataFrame({
    'ds': [
        s2d("2016-02-08"), s2d("2017-01-28"), s2d("2018-02-16"),
        s2d("2019-02-05"), s2d("2020-01-25")
    ],
    'holiday': "chinese_new_year",
    'lower_window': -20,
    'upper_window': 20,
})


def forecast(training_data, all_data):
    forecast = {}
    for c in all_data.keys():
        if (len(training_data) < 100):
            continue
        print("Starting with {}".format(c))
        model = Prophet(holidays=chinese_new_year)
        model.fit(training_data[c])
        forecast_period = model.make_future_dataframe(
            periods=(all_data[c].ds.max() - training_data[c].ds.max()).days,
            include_history=False
        )
        if (len(forecast_period) < 10):
            continue
        forecast[c] = model.predict(forecast_period)
        forecast[c]['ds'] = pd.to_datetime(forecast[c]['ds']).dt.date
        forecast[c] = forecast[c][["ds", "yhat", "yhat_lower", "yhat_upper"]]
        forecast[c] = forecast[c].merge(all_data[c], on="ds", how="inner")
        forecast[c]["delta"] = (forecast[c].y - forecast[c].yhat) / forecast[c].y
        forecast[c]["ci_delta"] = 0
        forecast[c].loc[
            forecast[c].delta > 0, "ci_delta"
        ] = forecast[c].delta / (forecast[c].yhat_upper - forecast[c].yhat)
        forecast[c].loc[
            forecast[c].delta < 0, "ci_delta"
        ] = forecast[c].delta / (forecast[c].yhat - forecast[c].yhat_lower)
        forecast[c] = forecast[c][["ds", "geo", "delta", "ci_delta"]]
        print("Done with {}".format(c))
    return forecast



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import datetime
import numpy as np


def add_days(date_string, n_days):
    """Add `n_days` days to a date string like '20190101'."""
    # TODO: would love to do some pre-computation here!
    original_date = datetime.datetime.strptime(date_string, '%Y%m%d')
    new_date = original_date + datetime.timedelta(days=n_days)
    return datetime.datetime.strftime(new_date, '%Y%m%d')


def filter_outliers(branch_data, threshold_quantile):
    """Return branch_data with outliers removed.

    N.B. `branch_data` is for an individual branch: if you do it for
    the entire experiment population in whole, then you may bias the
    results.

    TODO: here we remove outliers - should we have an option or
    default to cap them instead?

    Args:
        branch_data: Data for one branch as a 1D ndarray or similar.
        threshold_quantile (float): Discard outliers above this
            quantile.

    Returns:
        The subset of branch_data that was at or below the threshold
        quantile.
    """
    if threshold_quantile >= 1 or threshold_quantile < 0.5:
        raise ValueError("'threshold_quantile' should be close to 1")

    threshold_val = np.quantile(branch_data, threshold_quantile)

    return branch_data[branch_data <= threshold_val]



import pyspark.sql.functions as F

from pyspark.sql import Window


def get_consistency_check_quantities(self, enrollments, df):
    """Check that the enrollments view is consistent with the data view

    i.e.:
    - no missing client_ids from either
    - check for duplicate client_id values
    """
    return dict(
        dupe_client_id_enrollments=enrollments.filter(
            F.count(enrollments).over(Window.partitionBy(enrollments.client_id)) > 1
        ),

        missing_enrollments=df.filter(
            df.submission_date_s3 >= self.start_date
        ).filter(
            ~F.isnull(df.experiments[self.experiment_slug])
        ).select(
            df.client_id, df.experiments[self.experiment_slug]
        ).distinct(
        ).join(
            enrollments,
            [
                enrollments.client_id == df.client_id,
                enrollments.branch == df.experiments[self.experiment_slug],

                # TODO accuracy: once we can rely on
                #   `df.experiments[self.experiment_slug]`
                # existing even after unenrollment, we could start joining on
                # branch to reduce problems associated with split client_ids.

                # Do a quick pass aiming to efficiently filter out lots of rows:
                enrollments.enrollment_date <= df.submission_date_s3,

            ],
            how='left'
        ).filter(
            F.isnull(enrollments.enrollment_date)
        ).select(
            df.client_id,
            df.submission_date_s3,
            df.experiments[self.experiment_slug],
        ),
    )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import pyspark.sql.functions as F


def get_ad_ctr(sdf, groupers):
    """
    Calculates the sum of ad impressions, the sum of ad clicks
    and the click through rate (ctr) of ads for a given spark
    dataframe, grouped by a user-specificed set of columns

    Args
    sdf: a spark dataframe
    groupers: a list of colname names

    Returns a spark dataframe with the columns specified
    in the <groupers> arg along with:
    scalar_parent_browser_search_with_ads_sum,
    scalar_parent_browser_search_ad_clicks_sum,
    ctr
    """

    impressions_str = "scalar_parent_browser_search_with_ads"
    clicks_str = "scalar_parent_browser_search_ad_clicks"

    def explode_ad_map(metric):
        return (
          sdf
          .select("*", F.explode(metric))
          .groupby(groupers + [F.col("key").alias("engine")])
          .agg(F.coalesce(F.sum("value"), F.lit(0)).alias(metric + "_sum"))
        )

    impressions = explode_ad_map(impressions_str)
    clicks = explode_ad_map(clicks_str)

    return (
        impressions
        .join(clicks, on=groupers + ["engine"])
        .withColumn("ctr", F.col(clicks_str + '_sum') / F.col(impressions_str + '_sum'))
    )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


import pandas as pd
import numpy as np
import scipy.stats as stats
import statsmodels.stats.proportion as stp
from itertools import product
from tabulate import tabulate


def medianCI(data, ci, p):
    if type(data) is pd.Series or type(data) is pd.DataFrame:
        data = data.values

    data = data.reshape(-1)
    data = np.sort(data)
    N = data.shape[0]

    lowCount, upCount = stats.binom.interval(ci, N, p, loc=0)
    lowCount -= 1
    return data[int(lowCount)], data[int(upCount)]


def t_ci(x):
    df = len(x)-1
    scale = np.std(x) / np.sqrt(len(x))
    return stats.t.interval(alpha=0.95, df=df, loc=np.mean(x), scale=scale)


def mean_with_ci(x):
    return "{:.2f} ({:.2f} - {:.2f})".format(np.mean(x), t_ci(x)[0], t_ci(x)[1])


def median_with_ci(x):
    return "{:.2f} ({:.2f} - {:.2f})".format(
        np.median(x), medianCI(x, 0.95, 0.5)[0], medianCI(x, 0.95, 0.5)[1]
    )


def prop_greater_with_ci(x, thresh):
    num = np.sum(x >= thresh)
    den = float(len(x))
    ci = stp.proportion_confint(num, den)
    return "{:.2f} ({:.2f} - {:.2f})".format(num/den, ci[0], ci[1])


def gen_prop_greater_with_ci(thresh):
    return lambda x: prop_greater_with_ci(x, thresh)


class Metric:
    def __init__(self, name, function):
        self.name = name
        self.function = function


class SliceSpec:
    def __init__(self, slices={"All": lambda x: x}, slice_order=["All"]):
        self.slices = slices
        self.slice_order = slice_order

    @classmethod
    def complete_dimension(cls, data, var):
        return cls(
              {
                    "{}={}".format(var, val):
                    (lambda x, val=val: x.query("{}=='{}'".format(var, val)))
                    for val in data[var].unique()
              },
              ["{}={}".format(var, val) for val in sorted(data[var].unique())]
        )


def summary_table(data, vars, metrics, slice_spec=SliceSpec(), filter=None):
    table_data = [
        ["Variable"] + vars,
    ]
    table_data = table_data + [
        ["{} ({})".format(metric.name, slice)] + [
            (
                metric.function(
                    (
                        slice_spec.slices[slice](
                            data.dropna(subset=[var]).query(
                                filter.format(var=var)
                            )
                        )
                    )[var] if filter is not None else (
                        slice_spec.slices[slice](data.dropna(subset=[var]))
                    )[var]
                )
            ) for var in vars
        ]
        for (metric, slice) in product(metrics, slice_spec.slice_order)
    ]
    return tabulate(list(map(list, zip(*table_data))), headers="firstrow")



#!/usr/bin/env python
# encoding: utf-8

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import os.path
from setuptools import find_packages, setup

SETUP_PATH = os.path.dirname(os.path.abspath(__file__))


setup(
    name="dscontrib",
    author="Mozilla Data Science team",
    author_email="fx-data-dev@mozilla.org",
    description="A Python library for Mozilla Data Science code snippets",
    url="https://github.com/mozilla/dscontrib",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    install_requires=[
        "numpy",
        "pandas",
        "scipy",
        "mozanalysis"
    ],
    setup_requires=["setuptools_scm"],
    use_scm_version={"git_describe_command": os.path.join(SETUP_PATH, "describe_revision.py")},
    extras_require={
        "luigi": [
            "attrs",
            "luigi",
            "google-cloud-bigquery",
            "google-cloud-storage",
            "pytz",
        ],
    }
)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# This is a hack job to query an experiment in a mozanalysis-like way
# but with bigquery

import attr
import datetime
import re

from google.cloud import bigquery
from google.api_core.exceptions import Conflict


def add_days(date_string, n_days):
    """Add `n_days` days to a date string like '2019-01-01'."""
    original_date = datetime.datetime.strptime(date_string, '%Y-%m-%d')
    new_date = original_date + datetime.timedelta(days=n_days)
    return datetime.datetime.strftime(new_date, '%Y-%m-%d')


def date_sub(date_string_l, date_string_r):
    """Return the number of days between two date strings like '2019-01-01'"""
    date_l = datetime.datetime.strptime(date_string_l, '%Y-%m-%d')
    date_r = datetime.datetime.strptime(date_string_r, '%Y-%m-%d')
    return (date_l - date_r).days


def sanitize_table_name_for_bq(table_name):
    of_good_character_but_possibly_verbose = re.sub(r'[^a-zA-Z_0-9]', '_', table_name)

    if len(of_good_character_but_possibly_verbose) <= 1024:
        return of_good_character_but_possibly_verbose

    return of_good_character_but_possibly_verbose[:500] + '___' \
        + of_good_character_but_possibly_verbose[-500:]


class BigqueryStuff(object):
    def __init__(self, dataset_id, project_id='moz-fx-data-bq-data-science'):
        self.dataset_id = dataset_id
        self.project_id = project_id
        self.client = bigquery.Client(project=project_id)


def run_query(bq_stuff, sql, results_table=None):
    """Run a query and return the result.

    If ``results_table`` is provided, then save the results
    into there (or just query from there if it already exists).
    """
    if not results_table:
        return bq_stuff.client.query(sql).result()

    try:
        full_res = bq_stuff.client.query(
            sql,
            job_config=bigquery.QueryJobConfig(
                destination=bq_stuff.client.dataset(
                    bq_stuff.dataset_id
                ).table(results_table)
            )
        ).result()
        print('Saved into', results_table)
        return full_res

    except Conflict:
        print("Full results table already exists. Reusing", results_table)
        return bq_stuff.client.query(
            "SELECT * FROM `{project_id}.{dataset_id}.{full_table_name}`".format(
                project_id=bq_stuff.project_id,
                dataset_id=bq_stuff.dataset_id,
                full_table_name=results_table,
            )
        ).result()


@attr.s(frozen=True, slots=True)
class Experiment(object):
    """Query experiment data; store experiment metadata.

    The methods here query data in a way compatible with the following
    principles, which are important for experiment analysis:

    * The population of clients in each branch must have the same
      properties, aside from the intervention itself and its
      consequences; i.e. there must be no underlying bias in the
      branch populations.
    * We must measure the same thing for each client, to minimize the
      variance associated with our measurement.

    So that our analyses follow these abstract principles, we follow
    these rules:

    * Start with a list of all clients who enrolled.
    * We can filter this list of clients only based on information known
      to us at or before the time that they enrolled, because later
      information might be causally connected to the intervention.
    * For any given metric, every client gets a non-null value; we don't
      implicitly ignore anyone, even if they churned and stopped
      sending data.
    * Typically if an enrolled client no longer qualifies for enrollment,
      we'll still want to include their data in the analysis, unless
      we're explicitly using stats methods that handle censored data.
    * We define a "analysis window" with respect to clients'
      enrollment dates. Each metric only uses data collected inside
      this analysis window. We can only analyze data for a client
      if we have data covering their entire analysis window.


    Example usage (in a colab notebook)::

        from google.colab import auth
        auth.authenticate_user()
        print('Authenticated')

        project_id = 'moz-fx-data-bq-data-science'
        dataset_id = 'your-dataset-name'

        from mozanalysis.experiment import Experiment, BigqueryStuff
        from mozanalysis.metrics.desktop import active_hours, uri_count

        bq_stuff = BigqueryStuff(dataset_id, project_id)

        experiment = Experiment(
            experiment_slug='pref-fingerprinting-protections-retention-study-release-70',
            start_date='2019-10-29',
            num_dates_enrollment=8
        )

        # Run the query and store the results into a table
        res = experiment.get_single_window_data(
            enrollments,
            [
                active_hours,
                uri_count
            ],
            last_date_full_data='2019-12-01',
            analysis_start_days=0,
            analysis_length_days=7
        )

        # Pull data into a pandas df, ready for running stats
        pres = res.to_dataframe()

    Args:
        experiment_slug (str): Name of the study, used to identify
            the enrollment events specific to this study.
        start_date (str): e.g. '20190101'. First date on which enrollment
            events were received.
        num_dates_enrollment (int, optional): Only include this many dates
            of enrollments. If ``None`` then use the maximum number of dates
            as determined by the metric's analysis window and
            ``last_date_full_data``. Typically ``7n+1``, e.g. ``8``. The
            factor '7' removes weekly seasonality, and the ``+1`` accounts
            for the fact that enrollment typically starts a few hours
            before UTC midnight.

    Attributes:
        experiment_slug (str): Name of the study, used to identify
            the enrollment events specific to this study.
        start_date (str): e.g. '20190101'. First date on which enrollment
            events were received.
        num_dates_enrollment (int, optional): Only include this many days
            of enrollments. If ``None`` then use the maximum number of days
            as determined by the metric's analysis window and
            ``last_date_full_data``. Typically ``7n+1``, e.g. ``8``. The
            factor '7' removes weekly seasonality, and the ``+1`` accounts
            for the fact that enrollment typically starts a few hours
            before UTC midnight.
    """

    experiment_slug = attr.ib()
    start_date = attr.ib()
    num_dates_enrollment = attr.ib(default=None)

    def get_single_window_data(
        self, bq_stuff, metric_list, last_date_full_data,
        analysis_start_days, analysis_length_days, enrollments_query_type='normandy',
        custom_enrollments_query=None
    ):
        """Query per-client metric values.

        It will return the results, but it will also store them in a
        permanent table in BigQuery. The name of this table will be
        printed. Subsequent calls to this function will simply read
        the results from this table.

        Args:
            bq_stuff (BigqueryStuff): BigQuery configuration and client.
            metric_list (list of mozanalysis.metric.Metric): The metrics
                to analyze.
            last_date_full_data (str): The most recent date for which we
                have complete data, e.g. '20190322'. If you want to ignore
                all data collected after a certain date (e.g. when the
                experiment recipe was deactivated), then do that here.
            analysis_start_days (int): the start of the analysis window,
                measured in 'days since the client enrolled'. We ignore data
                collected outside this analysis window.
            analysis_length_days (int): the length of the analysis window,
                measured in days.
            enrollments_query_type (str): Specifies the query to use to
                get the experiment's enrollments, unless overridden by
                custom_enrollments_query.
            custom_enrollments_query (str): A full SQL query to be used
                in the main query::

                    WITH raw_enrollments AS ({custom_enrollments_query})

        Returns:
            A ``google.cloud.bigquery.table.RowIterator`` for the query
            results. What is that? I don't know either, but it is
            what you get when you construct a BigQuery ``query()`` then
            call ``.result()`` on it. You can call ``to_dataframe()``
            on this object to get a Pandas DataFrame.

            There is one row per ``client_id``. There are some metadata
            columns, then one column per metric in ``metric_list``,
            and one column per sanity-check metric.
            Columns (not necessarily in order):

                * client_id (str, optional): Not necessary for
                  "happy path" analyses.
                * branch (str): The client's branch
                * other columns of ``enrollments``.
                * [metric 1]: The client's value for the first metric in
                  ``metric_list``.
                * ...
                * [metric n]: The client's value for the nth (final)
                  metric in ``metric_list``.
                * [sanity check 1]: The client's value for the first
                  sanity check metric for the first data source that
                  supports sanity checks.
                * ...
                * [sanity check n]: The client's value for the last
                  sanity check metric for the last data source that
                  supports sanity checks.

            This format - the schema plus there being one row per
            enrolled client, regardless of whether the client has data
            in ``data_source`` - was agreed upon by the DS team, and is the
            standard format for queried experimental data.
        """
        time_limits = TimeLimits.for_single_analysis_window(
            self.start_date, last_date_full_data, analysis_start_days,
            analysis_length_days, self.num_dates_enrollment
        )

        full_sql = self._build_query(
            metric_list, time_limits, enrollments_query_type, custom_enrollments_query
        )

        full_res_table_name = sanitize_table_name_for_bq('_'.join(
            [last_date_full_data, self.experiment_slug, str(hash(full_sql))]
        ))

        return run_query(bq_stuff, full_sql, full_res_table_name)

    def get_time_series_data(
        self, bq_stuff, metric_list, last_date_full_data,
        time_series_period='weekly', enrollments_query_type='normandy',
        custom_enrollments_query=None
    ):
        """Query per-client metric values over a time series.

        Roughly equivalent to looping over ``get_single_window_data``
        with different analysis windows and reorganising the results.

        Args:
            bq_stuff (BigqueryStuff): BigQuery configuration and client.
            metric_list (list of mozanalysis.metric.Metric):
                The metrics to analyze.
            last_date_full_data (str): The most recent date for which we
                have complete data, e.g. '20190322'. If you want to ignore
                all data collected after a certain date (e.g. when the
                experiment recipe was deactivated), then do that here.
            time_series_period ('daily' or 'weekly'): How long each
                analysis window should be.
            enrollments_query_type (str): Specifies the query to use to
                get the experiment's enrollments, unless overridden by
                custom_enrollments_query.
            custom_enrollments_query (str): A full SQL query to be used
                in the main query::

                    WITH raw_enrollments AS ({custom_enrollments_query})

        Returns:
            A ``dict`` of data per analysis window. Each key is an ``int``:
            the number of days between enrollment and the start of the
            analysis window. Each value is a
            ``google.cloud.bigquery.table.RowIterator`` of results in "the
            standard format": one row per client, some metadata columns,
            plus one column per metric and sanity-check metric.
            Columns (not necessarily in order):

                * client_id (str, optional): Not necessary for
                  "happy path" analyses.
                * branch (str): The client's branch
                * other columns of ``enrollments``.
                * [metric 1]: The client's value for the first metric in
                  ``metric_list``.
                * ...
                * [metric n]: The client's value for the nth (final)
                  metric in ``metric_list``.
                * [sanity check 1]: The client's value for the first
                  sanity check metric for the first data source that
                  supports sanity checks.
                * ...
                * [sanity check n]: The client's value for the last
                  sanity check metric for the last data source that
                  supports sanity checks.

            Also returns a ``google.cloud.bigquery.table.RowIterator``
            for a table with _all_ results; one row per pair of client,
            analysis window.
        """

        time_limits = TimeLimits.for_ts(
            self.start_date, last_date_full_data, time_series_period,
            self.num_dates_enrollment
        )

        full_sql = self._build_query(
            metric_list, time_limits, enrollments_query_type, custom_enrollments_query
        )

        full_res_table_name = sanitize_table_name_for_bq('_'.join(
            [last_date_full_data, self.experiment_slug, str(hash(full_sql))]
        ))

        full_res = run_query(bq_stuff, full_sql, full_res_table_name)

        ts_res = {

            aw.start: run_query(
                bq_stuff,
                self._build_analysis_window_subset_query(
                    bq_stuff, aw, full_res_table_name
                ),
            )

            for aw in time_limits.analysis_windows
        }

        return ts_res, full_res

    def _build_query(
        self, metric_list, time_limits, enrollments_query_type,
        custom_enrollments_query=None,
    ):
        """Return SQL to query metric data for users.

        Building this query is the main goal of this module.
        """
        analysis_windows_query = self._build_analysis_windows_query(
            time_limits.analysis_windows
        )

        enrollments_query = custom_enrollments_query or \
            self._build_enrollments_query(time_limits, enrollments_query_type)

        metrics_columns, metrics_queries = self._build_metrics_query_bits(
            metric_list, time_limits
        )

        return """
    WITH analysis_windows AS (
        {analysis_windows_query}
    ),
    raw_enrollments AS ({enrollments_query}),
    enrollments AS (
        SELECT
            e.*,
            aw.*
        FROM raw_enrollments e
        CROSS JOIN analysis_windows aw
    )
    SELECT
        enrollments.*,
        {metrics_columns}
    FROM enrollments
    {metrics_queries}
        """.format(
            analysis_windows_query=analysis_windows_query,
            enrollments_query=enrollments_query,
            metrics_columns=',\n        '.join(metrics_columns),
            metrics_queries='\n'.join(metrics_queries)
        )

    @staticmethod
    def _build_analysis_window_subset_query(
        bq_stuff, analysis_window, full_res_table_name
    ):
        """Return SQL for partitioning time series results.

        When we query data for a time series, we query it for all
        points of the time series, and we store this in a table.

        This method returns SQL to query this table to obtain results
        in "the standard format" for a single analysis window.
        """
        return """
            SELECT * EXCEPT (client_id, analysis_window_start, analysis_window_end)
            FROM `{project_id}.{dataset_id}.{full_table_name}`
            WHERE analysis_window_start = {aws}
            AND analysis_window_end = {awe}
        """.format(
            project_id=bq_stuff.project_id,
            dataset_id=bq_stuff.dataset_id,
            full_table_name=full_res_table_name,
            aws=analysis_window.start,
            awe=analysis_window.end,
        )

    @staticmethod
    def _build_analysis_windows_query(analysis_windows):
        """Return SQL to construct a table of analysis windows.

        To query a time series, we construct a table of analysis windows
        and cross join it with the enrollments table to get one row per
        pair of client and analysis window.

        This method writes the SQL to define the analysis window table.
        """
        return "\n        UNION ALL\n        ".join(
            "(SELECT {aws} AS analysis_window_start, {awe} AS analysis_window_end)"
            .format(
                aws=aw.start,
                awe=aw.end,
            )
            for aw in analysis_windows
        )

    def _build_enrollments_query(self, time_limits, enrollments_query_type):
        """Return SQL to query a list of enrollments and their branches"""
        if enrollments_query_type == 'normandy':
            return self._build_enrollments_query_normandy(time_limits)
        elif enrollments_query_type == 'glean':
            raise NotImplementedError
        else:
            raise ValueError

    def _build_enrollments_query_normandy(self, time_limits):
        """Return SQL to query enrollments for a normandy experiment"""
        return """
        SELECT
            e.client_id,
            `moz-fx-data-shared-prod.udf.get_key`(e.event_map_values, 'branch')
                AS branch,
            min(e.submission_date) AS enrollment_date,
            count(e.submission_date) AS num_enrollment_events
        FROM
            `moz-fx-data-shared-prod.telemetry.events` e
        WHERE
            e.event_category = 'normandy'
            AND e.event_method = 'enroll'
            AND e.submission_date
                BETWEEN '{first_enrollment_date}' AND '{last_enrollment_date}'
            AND e.event_string_value = '{experiment_slug}'
        GROUP BY e.client_id, branch
            """.format(
            experiment_slug=self.experiment_slug,
            first_enrollment_date=time_limits.first_enrollment_date,
            last_enrollment_date=time_limits.last_enrollment_date,
        )

    def _build_metrics_query_bits(self, metric_list, time_limits):
        """Return lists of SQL fragments corresponding to metrics."""
        ds_metrics = self._partition_metrics_by_data_source(metric_list)

        metrics_columns = []
        metrics_queries = []

        for i, ds in enumerate(ds_metrics.keys()):
            query_for_metrics = ds.build_query(
                ds_metrics[ds], time_limits, self.experiment_slug
            )
            metrics_queries.append(
                """    LEFT JOIN (
        {query}
        ) ds_{i} USING (client_id, analysis_window_start, analysis_window_end)
                """.format(
                    query=query_for_metrics,
                    i=i
                )
            )

            for m in ds_metrics[ds]:
                metrics_columns.append("ds_{i}.{metric_name}".format(
                    i=i, metric_name=m.name
                ))

        return metrics_columns, metrics_queries

    def _partition_metrics_by_data_source(self, metric_list):
        """Return a dict mapping data sources to metric lists.

        Also add sanity metrics."""
        data_sources = {m.data_source for m in metric_list}

        return {
            ds:
                [m for m in metric_list if m.data_source == ds]
                + ds.get_sanity_metrics(self.experiment_slug)
            for ds in data_sources
        }


@attr.s(frozen=True, slots=True)
class DataSource(object):
    name = attr.ib(validator=attr.validators.instance_of(str))
    from_expr = attr.ib(validator=attr.validators.instance_of(str))
    experiments_column_type = attr.ib(default='desktop_main_ping', type=str)
    client_id_column = attr.ib(default='client_id', type=str)
    submission_date_column = attr.ib(default='submission_date', type=str)

    @property
    def experiments_column_expr(self):
        if self.experiments_column_type is None:
            return ''

        elif self.experiments_column_type == 'desktop_main_ping':
            return """AND (
                    ds.{submission_date} != e.enrollment_date
                    OR `moz-fx-data-shared-prod.udf.get_key`(
                        ds.experiments, '{experiment_slug}'
                    ) IS NOT NULL
                )"""

        elif self.experiments_column_type == 'glean':
            raise NotImplementedError

        else:
            raise ValueError

    def build_query(self, metric_list, time_limits, experiment_slug):
        return """SELECT
            e.client_id,
            e.analysis_window_start,
            e.analysis_window_end,
            {metrics}
        FROM enrollments e
            LEFT JOIN {from_expr} ds
                ON ds.{client_id} = e.client_id
                AND ds.{submission_date} BETWEEN '{fddr}' AND '{lddr}'
                AND ds.{submission_date} BETWEEN
                    DATE_ADD(e.enrollment_date, interval e.analysis_window_start day)
                    AND DATE_ADD(e.enrollment_date, interval e.analysis_window_end day)
                {ignore_pre_enroll_first_day}
        GROUP BY e.client_id, e.analysis_window_start, e.analysis_window_end""".format(
            client_id=self.client_id_column,
            submission_date=self.submission_date_column,
            from_expr=self.from_expr,
            fddr=time_limits.first_date_data_required,
            lddr=time_limits.last_date_data_required,
            metrics=',\n            '.join(
                "{se} AS {n}".format(se=m.select_expr, n=m.name)
                for m in metric_list
            ),
            ignore_pre_enroll_first_day=self.experiments_column_expr.format(
                submission_date=self.submission_date_column,
                experiment_slug=experiment_slug,
            )
        )

    def get_sanity_metrics(self, experiment_slug):
        if self.experiments_column_type is None:
            return []

        elif self.experiments_column_type == 'desktop_main_ping':
            return [
                Metric(
                    name=self.name + '_has_contradictory_branch',
                    data_source=self,
                    select_expr=agg_any("""`moz-fx-data-shared-prod.udf.get_key`(
                ds.experiments, '{experiment_slug}'
            ) != e.branch""".format(experiment_slug=experiment_slug)),
                ),
                Metric(
                    name=self.name + '_has_non_enrolled_data',
                    data_source=self,
                    select_expr=agg_any("""`moz-fx-data-shared-prod.udf.get_key`(
                ds.experiments, '{experiment_slug}'
            ) IS NULL""".format(experiment_slug=experiment_slug))
                ),
            ]

        elif self.experiments_column_type == 'glean':
            raise NotImplementedError

        else:
            raise ValueError


@attr.s(frozen=True, slots=True)
class Metric(object):
    name = attr.ib(type=str)
    data_source = attr.ib(type=DataSource)
    select_expr = attr.ib(type=str)


clients_daily = DataSource(
    name='clients_daily',
    from_expr="`moz-fx-data-shared-prod.telemetry.clients_daily`",
)

search_clients_daily = DataSource(
    name='search_clients_daily',
    from_expr='`moz-fx-data-shared-prod.search.search_clients_daily`',
    experiments_column_type=None,
)


def agg_sum(select_expr):
    return "COALESCE(SUM({}), 0)".format(select_expr)


def agg_any(select_expr):
    return "COALESCE(CAST(MAX({}) AS INT64), 0)".format(select_expr)


active_hours = Metric(
    name='active_hours',
    data_source=clients_daily,
    select_expr=agg_sum('active_hours_sum')
)

uri_count = Metric(
    name='uri_count',
    data_source=clients_daily,
    select_expr=agg_sum('scalar_parent_browser_engagement_total_uri_count_sum')
)

search_count = Metric(
    name='search_count',
    data_source=search_clients_daily,
    select_expr=agg_sum('sap')
)

ad_clicks = Metric(
    name='ad_clicks',
    data_source=search_clients_daily,
    select_expr=agg_sum('ad_click')
)


@attr.s(frozen=True, slots=True)
class TimeLimits(object):
    """Internal object containing various time limits.

    Instantiated and used by the ``Experiment`` class; end users
    should not need to interact with it.

    Do not directly instantiate: use the constructors provided.

    There are several time constraints needed to specify a valid query
    for experiment data:

        * When did enrollments start?
        * When did enrollments stop?
        * How long after enrollment does the analysis window start?
        * How long is the analysis window?

    Even if these four quantities are specified directly, it is
    important to check that they are consistent with the available
    data - i.e. that we have data for the entire analysis window for
    every enrollment.

    Furthermore, there are some extra quantities that are useful for
    writing efficient queries:

        * What is the first date for which we need data from our data
          source?
        * What is the last date for which we need data from our data
          source?

    Instances of this class store all these quantities and do validation
    to make sure that they're consistent. The "store lots of overlapping
    state and validate" strategy was chosen over "store minimal state
    and compute on the fly" because different state is supplied in
    different contexts.
    """

    first_enrollment_date = attr.ib(type=str)
    last_enrollment_date = attr.ib(type=str)

    first_date_data_required = attr.ib(type=str)
    last_date_data_required = attr.ib(type=str)

    analysis_windows = attr.ib()  # type: tuple[AnalysisWindow]

    @classmethod
    def for_single_analysis_window(
        cls,
        first_enrollment_date,
        last_date_full_data,
        analysis_start_days,
        analysis_length_dates,
        num_dates_enrollment=None,
    ):
        """Return a ``TimeLimits`` instance with the following parameters

        Args:
            first_enrollment_date (str): First date on which enrollment
                events were received; the start date of the experiment.
            last_date_full_data (str): The most recent date for which we
                have complete data, e.g. '20190322'. If you want to ignore
                all data collected after a certain date (e.g. when the
                experiment recipe was deactivated), then do that here.
            analysis_start_days (int): the start of the analysis window,
                measured in 'days since the client enrolled'. We ignore data
                collected outside this analysis window.
            analysis_length_days (int): the length of the analysis window,
                measured in days.
            num_dates_enrollment (int, optional): Only include this many days
                of enrollments. If ``None`` then use the maximum number of days
                as determined by the metric's analysis window and
                ``last_date_full_data``. Typically ``7n+1``, e.g. ``8``. The
                factor ``7`` removes weekly seasonality, and the ``+1``
                accounts for the fact that enrollment typically starts a few
                hours before UTC midnight.
        """
        analysis_window = AnalysisWindow(
            analysis_start_days, analysis_start_days + analysis_length_dates - 1
        )

        if num_dates_enrollment is None:
            last_enrollment_date = add_days(last_date_full_data, -analysis_window.end)

        else:
            last_enrollment_date = add_days(
                first_enrollment_date, num_dates_enrollment - 1
            )

            if add_days(
                last_enrollment_date, analysis_window.end
            ) > last_date_full_data:
                raise ValueError(
                    "You said you wanted {} dates of enrollment, ".format(
                        num_dates_enrollment
                    ) + "and need data from the {}th day after enrollment. ".format(
                        analysis_window.end
                    ) + "For that, you need to wait until we have data for {}.".format(
                        last_enrollment_date
                    )
                )

        first_date_data_required = add_days(
            first_enrollment_date, analysis_window.start
        )
        last_date_data_required = add_days(last_enrollment_date, analysis_window.end)

        tl = cls(
            first_enrollment_date=first_enrollment_date,
            last_enrollment_date=last_enrollment_date,
            first_date_data_required=first_date_data_required,
            last_date_data_required=last_date_data_required,
            analysis_windows=(analysis_window,),
        )
        return tl

    @classmethod
    def for_ts(
        cls,
        first_enrollment_date,
        last_date_full_data,
        time_series_period,
        num_dates_enrollment,
    ):
        """Return a ``TimeLimits`` instance for a time series.

        Args:
            first_enrollment_date (str): First date on which enrollment
                events were received; the start date of the experiment.
            last_date_full_data (str): The most recent date for which we
                have complete data, e.g. '20190322'. If you want to ignore
                all data collected after a certain date (e.g. when the
                experiment recipe was deactivated), then do that here.
            time_series_period: 'daily' or 'weekly'.
            num_dates_enrollment (int): Take this many days of client
                enrollments. This is a mandatory argument because it
                determines the number of points in the time series.
        """
        if time_series_period not in ('daily', 'weekly'):
            raise ValueError("Unsupported time series period {}".format(
                time_series_period
            ))

        analysis_window_length_dates = 1 if time_series_period == 'daily' else 7

        last_enrollment_date = add_days(
            first_enrollment_date, num_dates_enrollment - 1
        )
        max_dates_of_data = date_sub(last_date_full_data, last_enrollment_date) + 1
        num_periods = max_dates_of_data // analysis_window_length_dates

        if num_periods <= 0:
            raise ValueError("Insufficient data")

        analysis_windows = tuple([
            AnalysisWindow(
                i * analysis_window_length_dates,
                (i + 1) * analysis_window_length_dates - 1
            )
            for i in range(num_periods)
        ])

        last_date_data_required = add_days(
            last_enrollment_date, analysis_windows[-1].end
        )

        return cls(
            first_enrollment_date=first_enrollment_date,
            last_enrollment_date=last_enrollment_date,
            first_date_data_required=first_enrollment_date,
            last_date_data_required=last_date_data_required,
            analysis_windows=analysis_windows,
        )

    @first_enrollment_date.validator
    def _validate_first_enrollment_date(self, attribute, value):
        assert self.first_enrollment_date <= self.last_enrollment_date
        assert self.first_enrollment_date <= self.first_date_data_required
        assert self.first_enrollment_date <= self.last_date_data_required

    @last_enrollment_date.validator
    def _validate_last_enrollment_date(self, attribute, value):
        assert self.last_enrollment_date <= self.last_date_data_required

    @first_date_data_required.validator
    def _validate_first_date_data_required(self, attribute, value):
        assert self.first_date_data_required <= self.last_date_data_required

        min_analysis_window_start = min(aw.start for aw in self.analysis_windows)
        assert self.first_date_data_required == add_days(
            self.first_enrollment_date, min_analysis_window_start
        )

    @last_date_data_required.validator
    def _validate_last_date_data_required(self, attribute, value):
        max_analysis_window_end = max(aw.end for aw in self.analysis_windows)
        assert self.last_date_data_required == add_days(
            self.last_enrollment_date, max_analysis_window_end
        )


@attr.s(frozen=True, slots=True)
class AnalysisWindow(object):
    """Represents the range of days in which to measure a metric.

    The range is measured in "days after enrollment", and is inclusive.

    For example, ``AnalysisWindow(0, 6)`` is the first week after enrollment.

    Args:
        start (int): First day of the analysis window, in days since
            enrollment.
        end (int): Final day of the analysis window, in days since
            enrollment.
    """
    start = attr.ib(type=int)
    end = attr.ib(type=int)

    @start.validator
    def _validate_start(self, attribute, value):
        assert value >= 0

    @end.validator
    def _validate_end(self, attribute, value):
        assert value >= self.start



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import pandas as pd
import numpy as np
from datetime import timedelta
from collections import defaultdict
from plotly.offline import plot
import plotly.graph_objs as go
from dscontrib.jmccrosky.forecast.utils import matchDates, getLayout


def _getSinglePrediction(forecastData, asofdate, targetDate):
    temp = forecastData.query("asofdate == @asofdate & ds == @targetDate")
    if len(temp.yhat) == 1:
        return (temp.yhat.item(), temp.yhat_lower.item(), temp.yhat_upper.item())
    else:
        return None


def ValidateStability(forecastDataDict, asofdateRange, targetDate, suppressCI=False):
    data = {}
    for forecastDataKey in forecastDataDict:
        dates = []
        values = []
        for asofdate in asofdateRange:
            prediction = _getSinglePrediction(
                forecastDataDict[forecastDataKey], asofdate, targetDate)
            if prediction is not None:
                values.append([prediction])
                dates.append(asofdate)
        data[forecastDataKey] = pd.DataFrame({
            "date": dates,
            "Predicted MAU for {}".format(targetDate): [i[0][0] for i in values],
            "upper": [i[0][1] for i in values],
            "lower": [i[0][2] for i in values],
        })
    return plot(
        {
            "data": sum(([
                go.Scatter(
                    x=data[forecastDataKey]['date'],
                    y=data[forecastDataKey]["Predicted MAU for {}".format(targetDate)],
                    name="Prediction for {}".format(forecastDataKey),
                ),
                go.Scatter(
                    x=data[forecastDataKey]['date'],
                    y=data[forecastDataKey]['upper'],
                    fill='tonexty',
                    mode='none',
                    name='upper 80% CI for {}'.format(forecastDataKey),
                ) if not suppressCI else go.Scatter(),
                go.Scatter(
                    x=data[forecastDataKey]['date'],
                    y=data[forecastDataKey]['lower'],
                    fill='tonexty',
                    mode='none',
                    name='lower 80% CI for {}'.format(forecastDataKey),
                ) if not suppressCI else go.Scatter(),
            ] for forecastDataKey in forecastDataDict), []),
            "layout": getLayout(
                ("Predictions of MAU for {} using model "
                 "fit on data up to each date").format(targetDate),
                "Model Trained On Data Up To Date",
                "Prediction for {}".format(targetDate)
            )
        },
        output_type="div"
    )


def _getMetricForRange(actualData, forecastData, asofdate, metric):
    forecast = forecastData.query("asofdate == @asofdate & ds > @asofdate")
    matched = matchDates(
        actualData,
        forecast
    )
    metric = metric(
        np.array(matched.y),
        np.array(matched.yhat),
    )
    return metric


def ValidateMetric(actualData, forecastDataDict, asofdateRange, metric, metricName):
    data = {}
    for k in forecastDataDict:
        dates = []
        mapes = []
        for d in asofdateRange:
            mapes.append(_getMetricForRange(actualData, forecastDataDict[k], d, metric))
            dates.append(d)
        data[k] = pd.DataFrame({"date": dates, metricName: mapes})
    return plot(
        {
            "data": [go.Scatter(
                x=data[k]['date'],
                y=data[k][metricName],
                name="{} for {}".format(metricName, k)
            ) for k in forecastDataDict],
            "layout": getLayout(
                "{} for model trained up to date".format(metricName),
                "Model Trained On Data Up To Date",
                "{} Metric Value".format(metricName)
            )
        },
        output_type="div"
    )


def _getMetricTrace(model, data, trainingEndDate, metric, metricName):
    forecastStart = trainingEndDate + timedelta(days=1)
    forecastEnd = data.ds.max()
    model.fit(data.query("ds <= @trainingEndDate"))
    forecastPeriod = pd.DataFrame({'ds': pd.date_range(forecastStart, forecastEnd)})
    forecast = model.predict(forecastPeriod)
    matched = matchDates(
        data,
        forecast
    )
    return pd.DataFrame({
        "ds": matched.ds,
        metricName: [
            metric(
                np.array(matched.query("ds == @d").y),
                np.array(matched.query("ds == @d").yhat)
            )
            for d in matched.ds
        ]
    })


def ValidateTraces(modelGen, data, trainingEndDateRange, metric, metricName):
    traces = []
    for d in trainingEndDateRange:
        traces.append(_getMetricTrace(modelGen(), data, d, metric, metricName))
    return plot(
        {
            "data":
                [
                    go.Scatter(x=d['ds'], y=d[metricName])
                    for d in traces
                ],
                "layout": getLayout(
                    "Model Traces of {}".format(metricName),
                    "Prediction Date",
                    "{} Metric Value".format(metricName)
                )
        },
        output_type="div",
    )


def _accumulateHorizonMetrics(actualData, forecastData, asofdate, metric, metricValues):
    forecast = forecastData.query("asofdate == @asofdate & ds > @asofdate")
    matched = matchDates(
        actualData,
        forecast
    )
    for d in matched.ds:
        metricValue = metric(
            np.array(matched.query("ds == @d").y),
            np.array(matched.query("ds == @d").yhat)
        )
        horizon = (d - asofdate).days
        metricValues[horizon].append(metricValue)


def ValidateMetricHorizon(
    actualData, forecastDataDict, trainingEndDateRange, metric, metricName
):
    data = {}
    for k in forecastDataDict:
        metricValues = defaultdict(lambda: [])
        for d in trainingEndDateRange:
            _accumulateHorizonMetrics(
                actualData, forecastDataDict[k], d, metric, metricValues
            )
        data[k] = pd.DataFrame({
            "horizon": [i for i in metricValues.keys()],
            metricName: [np.mean(metricValues[h]) for h in metricValues.keys()]
        })
    return plot({
        "data": [
            go.Scatter(
                x=data[k]['horizon'],
                y=data[k][metricName],
                name="{} for {}".format(metricName, k)
            )
            for k in forecastDataDict
        ],
        "layout": getLayout(
            "{} Metric By Model Horizon".format(metricName),
            "Model Horizon",
            "Mean {} Metric Value".format(metricName)
        )
    }, output_type="div")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import pandas as pd


_kpi_queries = {
    "desktop": '''
        SELECT
            submission_date as date,
            sum(mau) AS global_mau,
            SUM(IF(country IN ('US', 'FR', 'DE', 'GB', 'CA'), mau, 0)) AS tier1_mau
        FROM
            `moz-fx-data-derived-datasets.telemetry.firefox_desktop_exact_mau28_by_dimensions_v1`
        GROUP BY
            date
        ORDER BY
            date
    ''',
    "nondesktop": '''
        SELECT
            submission_date as date,
            sum(mau) AS global_mau,
            SUM(IF(country IN ('US', 'FR', 'DE', 'GB', 'CA'), mau, 0)) AS tier1_mau
        FROM
            `moz-fx-data-derived-datasets.telemetry.firefox_nondesktop_exact_mau28_by_dimensions_v1`
        GROUP BY
            date
        ORDER BY
            date
    ''',
    "fxa": '''
        SELECT
            submission_date AS date,
            SUM(mau) AS global_mau,
            SUM(seen_in_tier1_country_mau) AS tier1_mau
        FROM
            `moz-fx-data-derived-datasets.telemetry.firefox_accounts_exact_mau28_by_dimensions_v1`
        GROUP BY
            submission_date
        ORDER BY
            submission_date
    '''
}


def getKPIData(bqClient, types=list(_kpi_queries.keys())):
    data = {}
    if not isinstance(types, list):
        types = [types]
    for q in types:
        rawData = bqClient.query(_kpi_queries[q]).to_dataframe()
        data['{}_global'.format(q)] = rawData[
            ["date", "global_mau"]
        ].rename(
            index=str, columns={"date": "ds", "global_mau": "y"}
        )
        data['{}_tier1'.format(q)] = rawData[
            ["date", "tier1_mau"]
        ].rename(
            index=str, columns={"date": "ds", "tier1_mau": "y"}
        )
    for k in data:
        data[k]['ds'] = pd.to_datetime(data[k]['ds']).dt.date
    return data


_nondesktop_query = '''
    SELECT
        submission_date as date,
        SUM(mau) AS global_mau,
        product
    FROM
        `moz-fx-data-derived-datasets.telemetry.firefox_nondesktop_exact_mau28_by_product_v1`
    GROUP BY
        submission_date,
        product
    ORDER BY
        date
    '''


def getNondesktopData(bqClient):
    data = {}
    rawData = bqClient.query(_nondesktop_query).to_dataframe()
    for p in [
        "Fennec Android", "Focus iOS", "Focus Android", "Fennec iOS", "Fenix",
        "Firefox Lite", "FirefoxForFireTV", "FirefoxConnect"
    ]:
        data['{}'.format(p)] = rawData.query("product == @p")[
            ["date", "global_mau"]
        ].rename(
            index=str, columns={"date": "ds", "global_mau": "y"}
        )
    for k in data:
        data[k]['ds'] = pd.to_datetime(data[k]['ds']).dt.date
    return data


_nondesktop_nofire_query = '''
    SELECT
        submission_date as date,
        SUM(mau) AS global_mau,
        SUM(tier1_mau) AS tier1_mau

    FROM
        `moz-fx-data-derived-datasets.telemetry.firefox_nondesktop_exact_mau28_by_product_v1`
    WHERE
        product != "FirefoxForFireTV"
    GROUP BY
        submission_date
    ORDER BY
        date
    '''


def getNondesktopNoFireData(bqClient):
    data = {}
    rawData = bqClient.query(_nondesktop_nofire_query).to_dataframe()
    data['nondesktop_nofire_global'] = rawData[
        ["date", "global_mau"]
    ].rename(
        index=str, columns={"date": "ds", "global_mau": "y"}
    )
    data['nondesktop_nofire_tier1'] = rawData[
        ["date", "tier1_mau"]
    ].rename(
        index=str, columns={"date": "ds", "tier1_mau": "y"}
    )
    for k in data:
        data[k]['ds'] = pd.to_datetime(data[k]['ds']).dt.date
    data['nondesktop_nofire_global_2020'] = data[
        'nondesktop_nofire_global'
    ].copy(deep=True)
    data['nondesktop_nofire_tier1_2020'] = data[
        'nondesktop_nofire_tier1'
    ].copy(deep=True)
    return data


_forecast_query = '''
    SELECT
        date,
        asofdate,
        mau,
        low90,
        high90
    FROM
        `{project}.{dataset}.{table}`
    WHERE
        datasource = '{product}'
        AND type = 'forecast'
        AND asofdate BETWEEN "{asofdateStart}" AND "{asofdateEnd}"
    ORDER BY
        date
    '''


def getForecastData(
    bqClient, project, dataset, table, product, asofdateStart, asofdateEnd
):
    rawData = bqClient.query(_forecast_query.format(
        project=project,
        dataset=dataset,
        table=table,
        product=product,
        asofdateStart=asofdateStart,
        asofdateEnd=asofdateEnd,
    )).to_dataframe().rename(columns={
        "date": "ds",
        "mau": "yhat",
        "high90": "yhat_upper",
        "low90": "yhat_lower",
    })
    rawData['ds'] = pd.to_datetime(rawData['ds']).dt.date
    return rawData



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import pandas as pd
import numpy as np
from google.cloud import bigquery
from google.cloud.exceptions import NotFound
from datetime import timedelta

from dscontrib.jmccrosky.forecast.models import setupModels, dataFilter


# Delete output table if necessary and create empty table with appropriate schema
def resetOuputTable(bigquery_client, project, dataset, table_name):
    dataset_ref = bigquery_client.dataset(dataset)
    table_ref = dataset_ref.table(table_name)
    try:
        bigquery_client.delete_table(table_ref)
    except NotFound:
        pass
    schema = [
        bigquery.SchemaField('asofdate', 'DATE', mode='REQUIRED'),
        bigquery.SchemaField('datasource', 'STRING', mode='REQUIRED'),
        bigquery.SchemaField('date', 'DATE', mode='REQUIRED'),
        bigquery.SchemaField('type', 'STRING', mode='REQUIRED'),
        bigquery.SchemaField('mau', 'FLOAT', mode='REQUIRED'),
        bigquery.SchemaField('low90', 'FLOAT', mode='REQUIRED'),
        bigquery.SchemaField('high90', 'FLOAT', mode='REQUIRED'),
    ] + [
        bigquery.SchemaField('p{}'.format(q), 'FLOAT', mode='REQUIRED')
        for q in range(10, 100, 10)
    ]
    table = bigquery.Table(table_ref, schema=schema)
    table = bigquery_client.create_table(table)
    return table


def writeForecasts(bigquery_client, table, modelDate, forecastEnd, data, product):
    minYear = data.ds.min().year
    maxYear = forecastEnd.year
    years = range(minYear, maxYear + 1)
    models = setupModels(years)
    forecastStart = modelDate + timedelta(days=1)
    forecastPeriod = pd.DataFrame({'ds': pd.date_range(forecastStart, forecastEnd)})
    data = dataFilter(data, product)
    models[product].fit(data.query("ds <= @modelDate"))
    forecastSamples = models[product].sample_posterior_predictive(
        models[product].setup_dataframe(forecastPeriod)
    )
    forecast = models[product].predict(forecastPeriod)
    outputData = {
        "asofdate": modelDate,
        "datasource": product,
        "date": forecast.ds,
        "type": "forecast",
        "mau": forecast.yhat,
        "low90": forecast.yhat_lower,
        "high90": forecast.yhat_upper,
    }
    outputData.update({
        "p{}".format(q): np.nanpercentile(forecastSamples['yhat'], q, axis=1)
        for q in range(10, 100, 10)
    })
    outputData = pd.DataFrame(outputData)[[
      "asofdate", "datasource", "date", "type", "mau", "low90", "high90",
      "p10", "p20", "p30", "p40", "p50", "p60", "p70", "p80", "p90"
    ]]
    outputData['date'] = pd.to_datetime(outputData['date']).dt.date
    errors = bigquery_client.insert_rows(
        table,
        list(outputData.itertuples(index=False, name=None))
    )
    assert errors == []
